
#pragma once

#include <core/base_column.hpp>
#include <iostream>

#include <utility>
#include <functional>
#include <algorithm>
//#include <vector>

//#include <unordered_map>
#include <boost/unordered_map.hpp>
#include <boost/any.hpp>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
//TBB includes
#include <tbb/parallel_sort.h>
#include "tbb/parallel_scan.h"
#include "base_table.hpp"
#include <core/foreign_key_constraint.hpp>
#include <tbb/task_scheduler_init.h>

#include <gpu/gpu_algorithms.hpp>

#include <util/time_measurement.hpp>
#include <util/begin_ptr.hpp>

//#include <core/column.hpp>

namespace CoGaDB{
/*!
 * 
 * 
 *  \brief     This class represents a column with type T, is the base class for all typed column classes and allows a uniform handling of columns of a certain type T. 
 *  \details   This class is indentended to be a base class, so it has a virtual destruktor and pure virtual methods, which need to be implemented in a derived class. 
 * 				Furthermore, it declares pure virtual methods to allow a generic handling of typed columns, e.g., operator[]. All algorithms can be applied to a typed 
 * 				column, because of this operator. This abstracts from a columns implementation detail, e.g., whether they are compressed or not. 
 *  \author    Sebastian Breß
 *  \version   0.2
 *  \date      2013
 *  \copyright GNU LESSER GENERAL PUBLIC LICENSE - Version 3, http://www.gnu.org/licenses/lgpl-3.0.txt
 */	
	
template<class T>
class ColumnBaseTyped : public ColumnBase{
	public:
        typedef boost::unordered_multimap<T,TID,boost::hash<T>, std::equal_to<T> > HashTable;
	//typedef boost::shared_ptr<ColumnBaseTyped> ColumnPtr;
	/***************** constructors and destructor *****************/
	ColumnBaseTyped(const std::string& name, AttributeType db_type);
	virtual ~ColumnBaseTyped();

	virtual bool insert(const boost::any& new_Value)=0;
	virtual bool insert(const T& new_Value) = 0;
	virtual bool update(TID tid, const boost::any& new_value) = 0;
	virtual bool update(PositionListPtr tid, const boost::any& new_value) = 0;	
	
	virtual bool remove(TID tid)=0;
	//assumes tid list is sorted ascending
	virtual bool remove(PositionListPtr tid)=0;
	virtual bool clearContent()=0;

	virtual const boost::any get(TID tid)=0;
        virtual std::string getStringValue(TID tid);
	//virtual const boost::any* const getRawData()=0;
	virtual void print() const throw()=0;
	virtual size_t size() const throw()=0;
	virtual unsigned int getSizeinBytes() const throw()=0;

	virtual const ColumnPtr copy() const=0;
        virtual const ColumnPtr gather(PositionListPtr tid_list) = 0; 
	/***************** relational operations on Columns which return lookup tables *****************/
	virtual const PositionListPtr sort(SortOrder order); 
	virtual const PositionListPtr selection(const boost::any& value_for_comparison, const ValueComparator comp);
        virtual const PositionListPtr selection(ColumnPtr, const ValueComparator comp);   
        virtual const BitmapPtr bitmap_selection(const boost::any& value_for_comparison, const ValueComparator comp);
	virtual const PositionListPtr parallel_selection(const boost::any& value_for_comparison, const ValueComparator comp, unsigned int number_of_threads);
        virtual const PositionListPtr lock_free_parallel_selection(const boost::any& value_for_comparison, const ValueComparator comp, unsigned int number_of_threads);
	//join algorithms
	virtual const PositionListPairPtr hash_join(ColumnPtr join_column);
        virtual const PositionListPairPtr parallel_hash_join(ColumnPtr join_column,unsigned int number_of_threads);
        static void hash_join_pruning_thread(ColumnBaseTyped<T>* join_column, HashTable* hashtable, unsigned int* join_tids_table1, unsigned int* join_tids_table2, unsigned int thread_id, unsigned int number_of_threads, unsigned int* result_size);
        static void join_write_result_chunk_thread(ColumnBaseTyped<T>* join_column, unsigned int* join_tids_table1, unsigned int* join_tids_table2, unsigned int* join_tids_result_table1, unsigned int* join_tids_result_table2, unsigned int thread_id, unsigned int number_of_threads, unsigned int begin_index_result, unsigned int end_index_result);
	virtual const PositionListPairPtr sort_merge_join(ColumnPtr join_column);
	virtual const PositionListPairPtr nested_loop_join(ColumnPtr join_column);
	virtual const PositionListPairPtr radix_join(ColumnPtr join_column);

	virtual bool add(const boost::any& new_Value);
	//vector addition between columns			
	virtual bool add(ColumnPtr join_column);

	virtual bool minus(const boost::any& new_Value);
	virtual bool minus(ColumnPtr join_column);	

	virtual bool multiply(const boost::any& new_Value);
	virtual bool multiply(ColumnPtr join_column);

	virtual bool division(const boost::any& new_Value);	
	virtual bool division(ColumnPtr join_column);	

	//template <typename U, typename BinaryOperator>
	//std::pair<ColumnPtr,ColumnPtr> aggregate_by_keys(ColumnBaseTyped<U>* keys, BinaryOperator binary_op) const;

	virtual bool store(const std::string& path) = 0;
	virtual bool load(const std::string& path) = 0;
	virtual bool isMaterialized() const  throw() = 0;
	virtual bool isCompressed() const  throw() = 0;
	virtual const ColumnPtr materialize() throw() =0;	
	virtual bool is_equal(ColumnPtr column);	
        virtual int compareValuesAtIndexes(TID id1, TID id2);	
        
        /* ATTENTION: after setting integrity constraints, the result of a write
         * operation is undefined! This follows the typical bulk load and 
         * analyize workflow of OLAP. Hence, we do currently not implement 
         * integrity checks when inserting new data. First import your data and then
         * set the constraints! */
        bool setPrimaryKeyConstraint();
        bool hasPrimaryKeyConstraint() const throw();
        
        bool hasForeignKeyConstraint() const throw();
        bool setForeignKeyConstraint(const ForeignKeyConstraint& prim_foreign_key_reference);
        const ForeignKeyConstraint& getForeignKeyConstraint();
        
	/*! \brief returns type information of internal values*/
	virtual const std::type_info& type() const throw();
	/*! \brief defines operator[] for this class, which enables the user to thread all typed columns as arrays.
	 * \details Note that this method is pure virtual, so it has to be defined in a derived class. 
	 * \return a reference to the value at position index
	 * */
	virtual T& operator[](const int index) = 0;
	inline bool operator==(ColumnBaseTyped<T>& column);
        protected:
        bool checkUniqueness();
        bool checkReferentialIntegrity(ColumnPtr primary_key_col);
        //private:
            bool has_primary_key_constraint_;
            bool has_foreign_key_constraint_;
            ForeignKeyConstraint fk_constr_;
//            //we use the names to avoid issues with serialization 
//            std::string name_of_referred_primary_key_col_;
//            std::string name_of_referred_primary_key_table_;
            
       
};


	template<class T>
	ColumnBaseTyped<T>::ColumnBaseTyped(const std::string& name, AttributeType db_type) : ColumnBase(name,db_type), has_primary_key_constraint_(false), has_foreign_key_constraint_(false),fk_constr_(){

	}

	template<class T>
	ColumnBaseTyped<T>::~ColumnBaseTyped(){

	}
        
	template<class T>
        std::string ColumnBaseTyped<T>::getStringValue(TID tid){
            std::stringstream ss;
            ss << (*this)[tid];
            return ss.str();
        }
        
template<class T>
const std::type_info& ColumnBaseTyped<T>::type() const throw(){
	return typeid(T);
}

	template<class T>
	bool ColumnBaseTyped<T>::is_equal(ColumnPtr column)
	{
		if(column->type()!=typeid(T)) {
			return false;
			//std::cout << "Fatal Error!!! Typemismatch for columns " << this->name_  << " and " << column->getName() << std::endl;
			//std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
			//exit(-1);
		}
		if(column->size()!=this->size())
			return false;
		shared_pointer_namespace::shared_ptr<ColumnBaseTyped<T> > typed_column = shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<T> >(column); //static_cast<IntColumnPtr>(column1);
		for(unsigned int i=0; i<this->size(); ++i) {
			if((*this)[i]!=(*typed_column)[i]) {
				return false;
			}
		}
		return true;
	}
        
	template<class T>
	int ColumnBaseTyped<T>::compareValuesAtIndexes(TID id1, TID id2){
            if ((*this)[id1] == (*this)[id2]) return 0;
            else if ((*this)[id1] < (*this)[id2]) return 1;
            else return -1;
            /*
 		if(column->type()!=typeid(T)) {
			return false;
		}
		assert(id_column<column->size() && id_this_col < this->size());
//			return false;
		shared_pointer_namespace::shared_ptr<ColumnBaseTyped<T> > typed_column = shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<T> >(column); //static_cast<IntColumnPtr>(column1);
		return (*this)[id_this_col]<typed_column[id_column]); */    
        }

        template<class T>
        bool ColumnBaseTyped<T>::setPrimaryKeyConstraint(){
//            std::set<T> value_set;     
//            for(TID i=0; i<this->size();++i){
//                value_set.insert((*this)[i]);
//            }
//            if(value_set.size()==this->size()){
//                has_primary_key_constraint_=true;
//                return true;
//            }else{
//                std::stringstream ss;
//                ss << "Cannot set Primary Key constraint: Column '" << this->getName() << "' contains duplicate keys!";
//                COGADB_ERROR(ss.str(),"");
//                return false;
//            }
            if(checkUniqueness()){
                has_primary_key_constraint_=true;
                return true;
            }else{
                has_primary_key_constraint_=false;
                return false;
            }
        }

        template<class T>
        bool ColumnBaseTyped<T>::hasPrimaryKeyConstraint() const throw(){
            return has_primary_key_constraint_;
        }
        
        template<class T>
        bool ColumnBaseTyped<T>::hasForeignKeyConstraint() const throw(){
            return has_foreign_key_constraint_;
        }
        template<class T>
        const ForeignKeyConstraint& ColumnBaseTyped<T>::getForeignKeyConstraint(){
            return this->fk_constr_;
        }
        template<class T>
        bool ColumnBaseTyped<T>::setForeignKeyConstraint(const ForeignKeyConstraint& foreign_key_constraint){
            //if we have already set a ForeignKeyConstraint, we cannot add a new one,
            //the old one needs to be deleted by the user!
            if(hasForeignKeyConstraint()) return false;
            ColumnPtr pk_column = foreign_key_constraint.getPrimaryKeyColumn();
            if(!pk_column) return false;
            //if(!pk_column->checkUniqueness()) return false;
            if(this->checkReferentialIntegrity(pk_column)){
                this->fk_constr_=foreign_key_constraint;
                this->has_foreign_key_constraint_=true;
                return true;
            }else{
                return false;
            }
        }
        template<class T>
        bool ColumnBaseTyped<T>::checkUniqueness(){
            std::set<T> value_set;     
            for(TID i=0; i<this->size();++i){
                value_set.insert((*this)[i]);
            }
            if(value_set.size()==this->size()){
                return true;
            }else{
                return false;
            }
        }
        template<class T>
        bool ColumnBaseTyped<T>::checkReferentialIntegrity(ColumnPtr primary_key_col){
            //if(primary_key_col->)
            if(primary_key_col->type()!=typeid(T)){
                COGADB_ERROR(std::string("Check Referential Integrity Failed! Typemismatch for columns ") + this->name_  + std::string(" and ") + primary_key_col->getName(),"");
                return false;
            }
//            if(primary_key_col->){
//                COGADB_ERROR(std::string("Check Referential Integrity Failed! Typemismatch for columns ") + this->name_  + std::string(" and ") + primary_key_col->getName(),"");
//                return false;
//            }
            shared_pointer_namespace::shared_ptr<ColumnBaseTyped<T> > pk_column = shared_pointer_namespace::dynamic_pointer_cast<ColumnBaseTyped<T> >(primary_key_col); //static_cast<IntColumnPtr>(column1);
            assert(pk_column!=NULL);
            if(!pk_column->hasPrimaryKeyConstraint()){
                COGADB_ERROR(std::string("Cannot create ForeignKey constrained: referenced column '") + primary_key_col->getName()  + std::string("' is not a primary key! "),"");
                return false;
            }
            if(!pk_column->checkUniqueness()){
                COGADB_ERROR(std::string("Column marked as PrimaryKey contains duplicates!: Column Name: '") + primary_key_col->getName() ,"");
                return false;
            }
            
	//perfrom simple hash join to check referential integrity
	typedef boost::unordered_multimap<T,TID,boost::hash<T>, std::equal_to<T> > HashTable;

	//create hash table
	HashTable hashtable;
	unsigned int hash_table_size=pk_column->size();
	unsigned int join_column_size=this->size();
        
	for(unsigned int i=0;i<hash_table_size;i++)	
		hashtable.insert(std::pair<T,TID> ((*pk_column)[i],i));

        std::pair<typename HashTable::iterator, typename HashTable::iterator> range;
        typename HashTable::iterator it;
	for(unsigned int i=0;i<join_column_size;i++){
                unsigned int number_of_matches=0;
		range =  hashtable.equal_range((*this)[i]);
		for(it=range.first ; it!=range.second;++it){
			if(it->first==(*this)[i]){ //(*join_column)[i]){
                           number_of_matches++;
			}
		}
                if(number_of_matches==0){
                    std::stringstream ss;
                    ss << "Cannot create ForeignKey constrained: broken reference: for value '" << (*this)[i] << "'! "; // << endl;
                    COGADB_ERROR(ss.str(),"");
                    return false;
                }
                if(number_of_matches>1){
                    std::stringstream ss;
                    ss << "Cannot create ForeignKey constrained: referenced column '" << primary_key_col->getName() << "' has duplicate value: '" << (*this)[i] << "'!"; 
                    COGADB_ERROR(ss.str(),"");
                    return false;
                }
	}
        return true;
            
            
            
//            for(unsigned int i=0;i<this->size();++i){
//                unsigned int number_of_matches=0;
//                for(unsigned int j=0;j<pk_column->size();++j){
//                    if((*pk_column)[j]==(*this)[i]){
//                         number_of_matches++;
//                    }
//                }
//                if(number_of_matches==0){
//                    std::stringstream ss;
//                    ss << "Cannot create ForeignKey constrained: broken reference: for value '" << (*this)[i] << "'! "; // << endl;
//                    COGADB_ERROR(ss.str(),"");
//                    return false;
//                }
//                if(number_of_matches>1){
//                    std::stringstream ss;
//                    ss << "Cannot create ForeignKey constrained: referenced column '" << primary_key_col->getName() << "' has duplicate value: '" << (*this)[i] << "'!"; 
//                    COGADB_ERROR(ss.str(),"");
//                    return false;
//                }
//            }
//            return true;
            
//            PositionListPairPtr tid_pairs = this->nested_loop_join(primary_key_col);
//            assert(tid_pairs->first->size(),tid_pairs->second->size());
//            //this is the foreign key column, if referential integrity is okay, 
//            //then the join results as many rows as our foreign key column 
//            if(tid_pairs->first->size()==this->size()){
//                
//            }
            
        }
        
        
template<class T>
const PositionListPtr ColumnBaseTyped<T>::sort(SortOrder order){

	PositionListPtr ids = PositionListPtr( new PositionList());
		std::vector<std::pair<T,TID> > v;

		for(unsigned int i=0;i<this->size();i++){
			v.push_back (std::pair<T,TID>((*this)[i],i) );
		}

		//TODO: change implementation, so that no copy operations are required -> use boost zip iterators!

	  	if(order==ASCENDING){
#ifdef FORCE_USE_UNSTABLE_SORT_FOR_BENCHMARK
			tbb::parallel_sort(v.begin(),v.end(),std::less_equal<std::pair<T,TID> >());
#else
	  		std::stable_sort(v.begin(),v.end(),std::less_equal<std::pair<T,TID> >());
#endif
		}else if(order==DESCENDING){
#ifdef FORCE_USE_UNSTABLE_SORT_FOR_BENCHMARK
			tbb::parallel_sort(v.begin(),v.end(),std::greater_equal<std::pair<T,TID> >());
#else
			std::stable_sort(v.begin(),v.end(),std::greater_equal<std::pair<T,TID> >()); 
#endif
		}else{
			std::cout << "FATAL ERROR: ColumnBaseTyped<T>::sort(): Unknown Sorting Order!" << std::endl;
		}

		for(unsigned int i=0;i<v.size();i++){
			ids->push_back(v[i].second);
		}

		return ids;
}

/*
		template<class T>
		struct Column_Scan_Equal : public unary_function<T,bool> {
			Column_Scan_Equal(std::vector<TID>& local_result_tids, const T& comparison_value) : local_result_tids_(local_result_tids), comparison_value_(comparison_value){}
  			bool operator() (T& value) {
				if(value==comparison_value_){
					local_result_tids_.push_back();
					return true;
				}
				return false;
			}
			std::vector<TID>& local_result_tids_;
			T comparison_value_;
		};*/

//template<class T>
//void selection_thread(unsigned int thread_id, unsigned int number_of_threads, const T& value, const ValueComparator comp, ColumnBaseTyped<T>* col, PositionListPtr result_tids){
//	//std::cout << "Hi I'm thread" << thread_id << std::endl;
//	if(!quiet) std::cout << "Using CPU for Selection (parallel mode)..." << std::endl;
//		unsigned int array_size=col->size();
//		for(TID i=thread_id;i<array_size;i+=number_of_threads){
//			
//			//boost::any value = column->get(i);
//			//val = values_[i];
//				
//			if(comp==EQUAL){
//                            if(value==(*col)[i]){
//                                    //result_table->insert(this->fetchTuple(i));
//                                    //std::cout << "MATCH " << thread_id << std::endl;
//                                    result_tids->push_back(i);
//                            }
//			}else if(comp==LESSER){
//                            if((*col)[i]<value){
//                                    //result_table->insert(this->fetchTuple(i));
//                                    //std::cout << "MATCH " << thread_id << std::endl;
//                                    result_tids->push_back(i);
//                            }
//			}else if(comp==LESSER_EQUAL){
//                            if((*col)[i]<=value){
//                                    //result_table->insert(this->fetchTuple(i));
//                                    //std::cout << "MATCH " << thread_id << std::endl;
//                                    result_tids->push_back(i);
//                            }
//			}else if(comp==GREATER){
//                            if((*col)[i]>value){
//                                    //std::cout << "MATCH " << thread_id << std::endl;
//                                    result_tids->push_back(i);
//                                    //result_table->insert(this->fetchTuple(i));
//                            }
//			}else if(comp==GREATER_EQUAL){
//                            if((*col)[i]>=value){
//                                    //std::cout << "MATCH " << thread_id << std::endl;
//                                    result_tids->push_back(i);
//                                    //result_table->insert(this->fetchTuple(i));
//                            }
//			}else{
//
//			}
//		}	
//}

//template<class T>
//const PositionListPtr ColumnBaseTyped<T>::parallel_selection(const boost::any& value_for_comparison, const ValueComparator comp, unsigned int number_of_threads){
//
//		PositionListPtr result_tids( new PositionList());
//		//unsigned int number_of_threads=4;
//
//		if(value_for_comparison.type()!=typeid(T)){
//			std::cout << "Fatal Error!!! Typemismatch for column " << name_ << std::endl;
//			std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
//			exit(-1);
//		}
//			
//		T value = boost::any_cast<T>(value_for_comparison);
//
//		boost::thread_group threads;
//
//		std::vector<PositionListPtr> local_result_arrays;
//		for(unsigned int i=0;i<number_of_threads;i++){
//			local_result_arrays.push_back(PositionListPtr(new std::vector<TID>()));
//			threads.add_thread(new boost::thread(boost::bind(&CoGaDB::selection_thread<T>, i, number_of_threads, value,  comp, this,local_result_arrays[i])));
//		}
//		threads.join_all();
//                
//                /*
//                unsigned int result_size=0;
//		for(unsigned int i=0;i<number_of_threads;i++){
//		//	std::cout << "Local result size of thread " << i << ":" << local_result_arrays[i]->size() << std::endl;
//			result_size+=local_result_arrays[i]->size();
//		}
//                result_tids = new PositionList(local_result_arrays[0]->begin(),local_result_arrays[0]->end()); //->resize(result_size);
//		for(unsigned int i=1;i<number_of_threads;i++){
//                        
//			std::merge (result_tids.begin(),result_tids.end(),local_result_arrays[i]->begin(),local_result_arrays[i]->end(),result_tids->begin());
//		}         */       
//                
//                
//                
//		for(unsigned int i=0;i<number_of_threads;i++){
//		//	std::cout << "Local result size of thread " << i << ":" << local_result_arrays[i]->size() << std::endl;
//			result_tids->insert(result_tids->end(),local_result_arrays[i]->begin(),local_result_arrays[i]->end());
//		}
//		//std::cout << "number of result tids:" << result_tids->size() << std::endl;
//		tbb::parallel_sort(result_tids->begin(),result_tids->end());
//		//exit(-1);
//		return result_tids;
//}

    typedef std::vector<int> Flags;
    typedef boost::shared_ptr<Flags> FlagsPtr;

    
    class TBB_Body_PrefixSum{
    public:
        TBB_Body_PrefixSum (std::vector<int>* y_, const std::vector<int>* x_) : sum(0), x(x_), y(y_) {}
        int get_sum() const {return sum;}
        template<typename Tag>
        void operator()( const tbb::blocked_range<int>& r, Tag ) {
            int temp = sum;
            for( int i=r.begin(); i<r.end(); ++i ) {
                temp = temp + (*x)[i];
                if(Tag::is_final_scan() )
                    (*y)[i] = temp;
            }
            sum = temp;
        }
        TBB_Body_PrefixSum ( TBB_Body_PrefixSum & b, tbb::split ) : x(b.x), y(b.y), sum(0) {}
        void reverse_join( TBB_Body_PrefixSum & a ) { sum = a.sum + sum;}
        void assign( TBB_Body_PrefixSum & b ) {sum = b.sum;}
    private:
        int sum;
        std::vector<int>* y;
        const std::vector<int>* x;
    };
    
    inline int TBB_Prefix_Sum(std::vector<int>& y, const std::vector<int>& x, unsigned int number_of_threads) {
        assert(y.size()==x.size());
        unsigned int chunk_size=x.size()/number_of_threads;
        TBB_Body_PrefixSum  body(&y,&x);
        tbb::parallel_scan( tbb::blocked_range<int>(0,x.size(),chunk_size), body );
        return body.get_sum();
    }

    
    
    
template <typename T>
void selection_thread_set_flag_array(ColumnBaseTyped<T>* col, unsigned int thread_id, FlagsPtr flags, unsigned int number_of_threads, const T& value_for_comparison, const ValueComparator comp){
        assert(flags->size()==col->size());
    	if(!quiet) std::cout << "Using CPU for Selection (parallel mode)..." << std::endl;
	unsigned int array_size=col->size();
        unsigned int chunk_size=col->size()/number_of_threads;
 	unsigned int start_id=thread_id*chunk_size;
   	unsigned int end_id=(thread_id*chunk_size)+chunk_size; 
        //make sure that the last thread processes the rest of the array
        if(thread_id+1==number_of_threads) end_id=array_size;
        if(comp==EQUAL){
            for(TID i=start_id;i<end_id;i++){
                if(value_for_comparison==(*col)[i]){
                    (*flags)[i]=1;
                }
            }
        }else if(comp==LESSER){
            for(TID i=start_id;i<end_id;i++){
                if((*col)[i]<value_for_comparison){
                    (*flags)[i]=1;
                }
            }
        }else if(comp==LESSER_EQUAL){
            for(TID i=start_id;i<end_id;i++){
                if((*col)[i]<=value_for_comparison){
                    (*flags)[i]=1;
                }
            }
        }else if(comp==GREATER){
            for(TID i=start_id;i<end_id;i++){
                if((*col)[i]>value_for_comparison){
                     (*flags)[i]=1;
                }
            }
        }else if(comp==GREATER_EQUAL){
            for(TID i=start_id;i<end_id;i++){
                if((*col)[i]>=value_for_comparison){
                     (*flags)[i]=1;
                }
            }            
        }else{
            std::cerr << "FATAL Error! In CoGaDB::selection_thread_set_flag_array(): Invalid Value Comparator! " << comp << std::endl;
        }
        
//	unsigned int array_size=col->size();   
//        if(comp==EQUAL){
//            for(TID i=thread_id;i<array_size;i+=number_of_threads){
//                if(value_for_comparison==(*col)[i]){
//                    (*flags)[i]=1;
//                }
//            }
//        }else if(comp==LESSER){
//            for(TID i=thread_id;i<array_size;i+=number_of_threads){
//                if((*col)[i]<value_for_comparison){
//                    (*flags)[i]=1;
//                }
//            }
//        }else if(comp==GREATER){
//            for(TID i=thread_id;i<array_size;i+=number_of_threads){
//                if((*col)[i]>value_for_comparison){
//                     (*flags)[i]=1;
//                }
//            }
//        }else{
//            std::cerr << "FATAL Error! In CoGaDB::selection_thread_set_flag_array(): Invalid Value Comparator! " << comp << std::endl;
//        }
                

}

template <typename T>
void selection_thread_write_result_to_output_array(ColumnBaseTyped<T>* col, unsigned int thread_id, FlagsPtr flags, std::vector<int>* prefix_sum_array, PositionListPtr result_tids, unsigned int number_of_threads){
        assert(flags->size()==col->size());
        assert(flags->size()==prefix_sum_array->size());
 	unsigned int array_size=col->size();
        unsigned int chunk_size=col->size()/number_of_threads;
 	unsigned int start_id=thread_id*chunk_size;
   	unsigned int end_id=(thread_id*chunk_size)+chunk_size; 
        //make sure that the last thread processes the rest of the array
        if(thread_id+1==number_of_threads) end_id=array_size;
        for(TID i=start_id;i<end_id;i++){
            if((*flags)[i]==1){
                unsigned int write_id=(*prefix_sum_array)[i]-1;
                (*result_tids)[write_id]=i; //write matching TID to output buffer
            }
        }

    	//if(!quiet) std::cout << "Using CPU for Selection (parallel mode)..." << std::endl;
//		unsigned int array_size=col->size();
//		for(TID i=thread_id;i<array_size;i+=number_of_threads){
//                    if((*flags)[i]==1){
//                        unsigned int write_id=(*prefix_sum_array)[i]-1;
//                        (*result_tids)[write_id]=i; //write matching TID to output buffer
//                    }
//                }

}


template<class T>
const PositionListPtr ColumnBaseTyped<T>::lock_free_parallel_selection(const boost::any& value_for_comparison, const ValueComparator comp, unsigned int number_of_threads){

		
		//unsigned int number_of_threads=4;

		if(value_for_comparison.type()!=typeid(T)){
			std::cout << "Fatal Error!!! Typemismatch for column " << name_ << std::endl;
			std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
			exit(-1);
		}
			
		T value = boost::any_cast<T>(value_for_comparison);

		boost::thread_group threads;

                //create flag array of column size and init with zeros 
                FlagsPtr flags(new Flags(this->size(),0)); 
		//std::vector<PositionListPtr> local_result_arrays;
		for(unsigned int i=0;i<number_of_threads;i++){
			threads.add_thread(new boost::thread(boost::bind(&CoGaDB::selection_thread_set_flag_array<T>, this, i, flags, number_of_threads, value,  comp)));
		}
		threads.join_all();
                
                if(!quiet && verbose && debug)
                {
                    std::cout << "FLAG Array:" << std::endl;
                    for(unsigned int i=0;i<flags->size();++i){
                        std::cout << (*flags)[i] << std::endl;
                    }
                }

                std::vector<int> prefix_sum(this->size(),0);
                
                //do prefix sum on threads
                TBB_Prefix_Sum(prefix_sum,*flags,number_of_threads);
                
                if(!quiet && verbose && debug)
                {
                    std::cout << "Prefix Sum:" << std::endl;
                    for(unsigned int i=0;i<prefix_sum.size();++i){
                        std::cout << prefix_sum[i] << std::endl;
                    }
                }
                unsigned int resul_size=prefix_sum.back();
                PositionListPtr result_tids( new PositionList(resul_size));
                
 		for(unsigned int i=0;i<number_of_threads;i++){
			threads.add_thread(new boost::thread(boost::bind(&CoGaDB::selection_thread_write_result_to_output_array<T>, this, i, flags, &prefix_sum, result_tids, number_of_threads)));
		}
		threads.join_all();
                
                if(!quiet && verbose && debug)
                {
                    std::cout << "TIDS:" << std::endl;
                    for(unsigned int i=0;i<result_tids->size();++i){
                        std::cout << (*result_tids)[i] << std::endl;
                    }
                }
                
		return result_tids;
}



template<class T>
const PositionListPtr ColumnBaseTyped<T>::selection(const boost::any& value_for_comparison, const ValueComparator comp){
    if(value_for_comparison.type()!=typeid(T)){
            std::cout << "Fatal Error!!! Typemismatch for column " << name_ << std::endl;
            std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
            exit(-1);
    }

    T value = boost::any_cast<T>(value_for_comparison);
    PositionListPtr result_tids;

    result_tids = PositionListPtr(new PositionList());
    //one third rule for selections: assume a selectivity of 0.3, meaning we need roughly 0.3 times of the input to store the result 
    //this optimizatio nshould minimize the number of reallocations during the insertion process
    //result_tids->reserve(0.3*this->size()); 
    
    //std::cout << "Scan column of size " << this->size() << std::endl; 
    unsigned int array_size=this->size();
//    //calls new internally   
//     result_tids->resize(array_size);
//     //tids.reserve(array_size);
//     //get pointer
//     unsigned int* array_tids=hype::util::begin_ptr(*result_tids);
//     assert(array_tids!=NULL);
//     unsigned int pos=0;
    if(!quiet) std::cout << "Using CPU for Selection..." << std::endl;
    //unsigned int array_size = this->size();
    //for(TID i=0;i<array_size;i++){

            //boost::any value = column->get(i);
            //val = values_[i];

//        unsigned int pos =0;
//        for(TID i=0;i<array_size;i++){
//                    result_tids[pos]=i;
//                    pos+=(value==(*this)[i]);
//        } 
    
    if(comp==EQUAL){
        for(TID i=0;i<array_size;i++){
            if(value==(*this)[i]){
                    //array_tids[pos++]=i;
                    result_tids->push_back(i);
            }
        }
    }else if(comp==LESSER){
        for(TID i=0;i<array_size;i++){
            if((*this)[i]<value){
                    //result_table->insert(this->fetchTuple(i));
                    result_tids->push_back(i);
                    //array_tids[pos++]=i;
            }
        }
    }else if(comp==LESSER_EQUAL){
        for(TID i=0;i<array_size;i++){
            if((*this)[i]<=value){
                    result_tids->push_back(i);
                    //result_table->insert(this->fetchTuple(i));
                    //array_tids[pos++]=i;
            }
        } 
    }else if(comp==GREATER){
        for(TID i=0;i<array_size;i++){
            if((*this)[i]>value){
                    result_tids->push_back(i);
                    //result_table->insert(this->fetchTuple(i));
                    //array_tids[pos++]=i;
            }
        }
    }else if(comp==GREATER_EQUAL){
        for(TID i=0;i<array_size;i++){
            if((*this)[i]>=value){
                    result_tids->push_back(i);
                    //result_table->insert(this->fetchTuple(i));
                    //array_tids[pos++]=i;
            }
        }          
    }else{

    }
    //}
    //shrink to actual result size
    //result_tids->resize(pos);
    return result_tids;
}

template<class T>
const BitmapPtr ColumnBaseTyped<T>::bitmap_selection(const boost::any& value_for_comparison, const ValueComparator comp){
    COGADB_FATAL_ERROR("Called function bitmap_selection of class ColumnBaseTyped<T>!","");
    return BitmapPtr();
}


template<class T>
const PositionListPtr ColumnBaseTyped<T>::selection(ColumnPtr comparison_column, const ValueComparator comp){
    assert(comparison_column!=NULL);
    if(comparison_column->type()!=typeid(T)){
            std::cout << "Fatal Error!!! Typemismatch for columns " << this->name_  << " and " << comparison_column->getName() << std::endl;
            std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
            exit(-1);
    }

    shared_pointer_namespace::shared_ptr<ColumnBaseTyped<T> > column = shared_pointer_namespace::dynamic_pointer_cast<ColumnBaseTyped<T> >(comparison_column); //static_cast<IntColumnPtr>(column1);
    assert(column!=NULL);
    
    PositionListPtr result_tids;

    result_tids = PositionListPtr(new PositionList());
    //one third rule for selections: assume a selectivity of 0.3, meaning we need roughly 0.3 times of the input to store the result 
    //this optimizatio nshould minimize the number of reallocations during the insertion process
    result_tids->reserve(this->size()); 
    if(!quiet) std::cout << "Using CPU for Selection..." << std::endl;
    unsigned int array_size = this->size();
    //for(TID i=0;i<array_size;i++){

            //boost::any value = column->get(i);
            //val = values_[i];

    if(comp==EQUAL){
        for(TID i=0;i<array_size;i++){
            if((*column)[i]==(*this)[i]){
                    result_tids->push_back(i);
            }
        }
    }else if(comp==LESSER){
        for(TID i=0;i<array_size;i++){
            if((*this)[i]<(*column)[i]){
                    //result_table->insert(this->fetchTuple(i));
                    result_tids->push_back(i);
            }
        }
    }else if(comp==LESSER_EQUAL){
        for(TID i=0;i<array_size;i++){
            if((*this)[i]<=(*column)[i]){
                    //result_table->insert(this->fetchTuple(i));
                    result_tids->push_back(i);
            }
        }
    }else if(comp==GREATER){
        for(TID i=0;i<array_size;i++){
            if((*this)[i]>(*column)[i]){
                    result_tids->push_back(i);
                    //result_table->insert(this->fetchTuple(i));
            }
        }
    }else if(comp==GREATER_EQUAL){
        for(TID i=0;i<array_size;i++){
            if((*this)[i]>=(*column)[i]){
                    result_tids->push_back(i);
                    //result_table->insert(this->fetchTuple(i));
            }
        }
    }else{

    }
    //}	

    return result_tids;
}

	
	template<class T>
	const PositionListPairPtr ColumnBaseTyped<T>::hash_join(ColumnPtr join_column_){

		typedef boost::unordered_multimap<T,TID,boost::hash<T>, std::equal_to<T> > HashTable;

				if(join_column_->type()!=typeid(T)){
					std::cout << "Fatal Error!!! Typemismatch for columns " << this->name_  << " and " << join_column_->getName() << std::endl;
					std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
					exit(-1);
				}
				
				shared_pointer_namespace::shared_ptr<ColumnBaseTyped<T> > join_column = shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<T> >(join_column_); //static_cast<IntColumnPtr>(column1);

				PositionListPairPtr join_tids( new PositionListPair());
				join_tids->first = PositionListPtr( new PositionList() );
				join_tids->second = PositionListPtr( new PositionList() );

                                
        Timestamp build_hashtable_begin = getTimestamp();
	//create hash table
	HashTable hashtable;
	unsigned int hash_table_size=this->size();
	unsigned int join_column_size=join_column->size();

        assert(join_column_->size()>=this->size());                        
//        unsigned int* join_tids_table1 =  new unsigned int[join_column_size];   
//        unsigned int* join_tids_table2 =  new unsigned int[join_column_size]; 
        
          unsigned int* join_tids_table1 = (unsigned int*) malloc(join_column_size*sizeof(T));   
          unsigned int* join_tids_table2 = (unsigned int*) malloc(join_column_size*sizeof(T));
        
//        unsigned int pos1 = 0;
//        unsigned int pos2 = 0;
        unsigned int pos = 0;
        ColumnBaseTyped<T>& join_column_ref = dynamic_cast< ColumnBaseTyped<T>& >(*join_column);
        
	for(unsigned int i=0;i<hash_table_size;i++)	
		hashtable.insert(std::pair<T,TID> ((*this)[i],i));
        Timestamp build_hashtable_end = getTimestamp();
//        std::cout << "Number of Buckets: " << hashtable.bucket_count() << std::endl;
//        for(unsigned int i=0;i< hashtable.bucket_count();i++){
//            std::cout << "Size of Bucket '" << i << "': " << hashtable.bucket_size(i) << std::endl;
//        }
        
	//probe larger relation
        Timestamp prune_hashtable_begin = getTimestamp();

        std::pair<typename HashTable::iterator, typename HashTable::iterator> range;
        typename HashTable::iterator it;
	for(unsigned int i=0;i<join_column_size;i++){
		range =  hashtable.equal_range(join_column_ref[i]);
		for(it=range.first ; it!=range.second;++it){
			if(it->first==join_column_ref[i]){ //(*join_column)[i]){
                                  join_tids_table1[pos]=it->second;
                                  join_tids_table2[pos]=i;
                                  pos++;
                                  //pos2=++pos1;
//                                join_tids_table1[pos1++]=it->second;
//                                join_tids_table2[pos2++]=i;
//				join_tids->first->push_back(it->second);
//				join_tids->second->push_back(i);
                                
				//cout << "match! " << it->second << ", " << i << "	"  << it->first << endl;
			}
		}
	}
        
//        for (unsigned int i = 0; i < join_column_size; i++) {
//            std::pair<typename HashTable::iterator, typename HashTable::iterator> range = hashtable.equal_range((*join_column)[i]);
//            for (typename HashTable::iterator it = range.first; it != range.second; ++it) {
//                if (it->first == join_column_ref[i]) { //(*join_column)[i]){
//                    join_tids_table1[pos1++] = it->second;
//                    join_tids_table2[pos2++] = i;
//                    //				join_tids->first->push_back(it->second);
//                    //				join_tids->second->push_back(i);
//
//                    //cout << "match! " << it->second << ", " << i << "	"  << it->first << endl;
//                }
//            }
//        }
        //copy result in PositionList (vector)
        join_tids->first->insert(join_tids->first->end(),join_tids_table1,join_tids_table1+pos);
        join_tids->second->insert(join_tids->second->end(),join_tids_table2,join_tids_table2+pos);
        
//        delete join_tids_table1;
//        delete join_tids_table2;
        free(join_tids_table1);
        free(join_tids_table2);      
        Timestamp prune_hashtable_end = getTimestamp();

	if(!quiet && verbose)
		std::cout << "Hash Join: Build Phase: " << double(build_hashtable_end-build_hashtable_begin)/(1000*1000) << "ms"
		          << "Pruning Phase: "  << double(prune_hashtable_end-prune_hashtable_begin)/(1000*1000) << "ms" << std::endl; 
        
		return join_tids;
	}


        
        
//template <typename T> template< typename U, typename BinaryOperator>
//std::pair<ColumnPtr,ColumnPtr> ColumnBaseTyped<T>::aggregate_by_keys(ColumnBaseTyped<U>* keys, BinaryOperator binary_op) const{

//				/*
//				if(keys->type()!=typeid(T)){
//					std::cout << "Fatal Error!!! Typemismatch for columns " << this->name_  << " and " << keys->getName() << std::endl;
//					std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
//					exit(-1);
//				}*/
// 
//				assert(keys->size()==this->size());

// Column<T>* new_keys = new Column<T>(this->name_,this->getType());
// Column<T>* aggregated_values = new Column<T>(this->name_,this->getType()); 

// T val=0;
// for(unsigned int i=0;i<this->size();i++){
//   
//  if(i==0){
//   val=(*keys)[i];
//   continue;
//  }

// if(  (*keys)[i-1]==(*keys)[i] ){
//   val = binary_op(val,(*this)[i]);  
// }else{
//  new_keys->insert((*keys)[i]);
//  aggregated_values->insert(val);
//  val=(*keys)[i];
// }

// }

// return std::pair<ColumnPtr,ColumnPtr>(ColumnPtr(new_keys), ColumnPtr(aggregated_values));
////return std::pair<ColumnPtr,ColumnPtr>(ColumnPtr(), ColumnPtr());
//}


	template<class Type>
	const PositionListPairPtr ColumnBaseTyped<Type>::sort_merge_join(ColumnPtr join_column_){

				if(join_column_->type()!=typeid(Type)){
					std::cout << "Fatal Error!!! Typemismatch for columns " << this->name_  << " and " << join_column_->getName() << std::endl;
					std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
					exit(-1);
				}
				
				shared_pointer_namespace::shared_ptr<ColumnBaseTyped<Type> > join_column = shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<Type> >(join_column_); //static_cast<IntColumnPtr>(column1);

				PositionListPairPtr join_tids( new PositionListPair());
				join_tids->first = PositionListPtr( new PositionList() );
				join_tids->second = PositionListPtr( new PositionList() );

//				if(comp_dev==GPU){
//					std::cout << "No GPU Sort Merge Join implemented!" << std::endl;
//					exit(-1);
//				}

				PositionListPtr tids_sorted_join_column1 = this->sort(ASCENDING); //doesnt change column itself, only creates lookup table
				PositionListPtr tids_sorted_join_column2 = join_column->sort(ASCENDING);

				assert(tids_sorted_join_column1!=NULL);
//				assert(tids_sorted_join_column1->first!=NULL);
//				assert(tids_sorted_join_column1->second!=NULL);

				assert(tids_sorted_join_column2!=NULL);

				assert(tids_sorted_join_column1->size()==this->size());
				assert(tids_sorted_join_column2->size()==join_column->size());

				TID i=0;
				TID j=0;
				bool end=false;
				while(!end){
						if(!quiet && verbose && debug){
							std::cout << "index i: " << i << "(max: " << tids_sorted_join_column1->size() 
							     << ") index j:" << j << "(max: " << tids_sorted_join_column2->size() << ")" << std::endl;
						}
						Type val = (*this)[(*tids_sorted_join_column1)[i]];
						Type val2 = (*(join_column))[(*tids_sorted_join_column2)[j]]; //j);	
						if(val>val2){
							j++;
							if(j>=join_column->size()-1){
							  end=true;
								
							}
						}else if (val<val2){
							i++;
							if(i>=this->size()-1){
							   end=true;
							}
						}else if(val==val2){
								Type tmp_val=val;
								unsigned int index=0;
								while(tmp_val==val){
									if(i+index==this->size()-1){ index++; break;} 
									index++;
									tmp_val = (*this)[(*tids_sorted_join_column1)[i+index]];
									//tmp_val = any_cast<T>(tmp_value);
								}
								tmp_val=val2;
								unsigned int index2=0;
								while(tmp_val==val2){
				
									if(j+index2==join_column->size()-1){ index2++; break;} 
									index2++;
									tmp_val = (*join_column)[(*tids_sorted_join_column2)[j+index2]];
									//tmp_val = any_cast<T>(tmp_value);
								}								
								for(unsigned int k=i;k<i+index;k++){
									for(unsigned int l=j;l<j+index2;l++){
//										join_tids.push_back(TID_Pair(tids_sorted_join_column1[k],tids_sorted_join_column2[l]));
										join_tids->first->push_back((*tids_sorted_join_column1)[k]);
										join_tids->second->push_back((*tids_sorted_join_column2)[l]);
									}
								}

								i=i+index;
								j=j+index2;

								if(i>=this->size()-1 || j>=join_column->size()-1){
									end=true;
								}

						}
				}
				return join_tids;
	}


	template<class Type>
	const PositionListPairPtr ColumnBaseTyped<Type>::nested_loop_join(ColumnPtr join_column_){
			assert(join_column_!=NULL);
			if(join_column_->type()!=typeid(Type)){
				std::cout << "Fatal Error!!! Typemismatch for columns " << this->name_  << " and " << join_column_->getName() << std::endl;
				std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
				exit(-1);
			}
			
			shared_pointer_namespace::shared_ptr<ColumnBaseTyped<Type> > join_column = shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<Type> >(join_column_); //static_cast<IntColumnPtr>(column1);

			PositionListPairPtr join_tids( new PositionListPair());
			join_tids->first = PositionListPtr( new PositionList() );
			join_tids->second = PositionListPtr( new PositionList() );


			unsigned int join_column1_size=this->size();
			unsigned int join_column2_size=join_column->size();	

			for(unsigned int i=0;i<join_column1_size;i++){
				for(unsigned int j=0;j<join_column2_size;j++){
					if((*this)[i]==(*join_column)[j]){
						if(debug) std::cout << "MATCH: (" << i << "," << j << ")" << std::endl;
						join_tids->first->push_back(i);
						join_tids->second->push_back(j);
					}
				}
			}

		

		return join_tids;
	}

        
	template<class Type>
	const PositionListPairPtr ColumnBaseTyped<Type>::radix_join(ColumnPtr join_column){
            return this->hash_join(join_column);
        }
        
	template<class T>
	bool ColumnBaseTyped<T>::operator==(ColumnBaseTyped<T>& column){
		if(this->size()!=column.size()) return false;
		for(unsigned int i=0;i<this->size();i++){
			if((*this)[i]!=column[i]){ 	
				return false;
			}
		}
		return true;
	}

	template<class Type>
	bool ColumnBaseTyped<Type>::add(const boost::any& new_value){
		if(new_value.empty()) return false;
		if(typeid(Type)==new_value.type()){
			 Type value = boost::any_cast<Type>(new_value);
			 //std::transform(myvec.begin(), myvec.end(), myvec.begin(),
          //bind2nd(std::plus<double>(), 1.0));
			 for(unsigned int i=0;i<this->size();i++){
					this->operator[](i)+=value;
			 }
			 return true;
		}
		return false;
	}
	

			
	template<class Type>
	bool ColumnBaseTyped<Type>::add(ColumnPtr column){
		//std::transform ( first, first+5, second, results, std::plus<int>() );		
		shared_pointer_namespace::shared_ptr<ColumnBaseTyped<Type> > typed_column = shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<Type> >(column);
		if(!column) return false;
		for(unsigned int i=0;i<this->size();i++){
			this->operator[](i)+=typed_column->operator[](i);
		}			
		return true;
	}
/*
	template<class Type>
	bool ColumnBaseTyped<Type>::add(ColumnPtr column){
		//std::transform ( first, first+5, second, results, std::plus<int>() );		
		shared_pointer_namespace::shared_ptr<ColumnBaseTyped<Type> > typed_column = shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<Type> >(column);
		if(!column) return false;
		for(unsigned int i=0;i<this->size();i++){
			this->operator[](i)+=typed_column->operator[](i);
		}			
		return true;
	}*/

	template<class Type>
	bool ColumnBaseTyped<Type>::minus(const boost::any& new_value){
		//shared_pointer_namespace::shared_ptr<ColumnBaseTyped<Type> > typed_column = shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<Type> >(column);	
		if(new_value.empty()) return false;
		if(typeid(Type)==new_value.type()){
			 Type value = boost::any_cast<Type>(new_value);
			 for(unsigned int i=0;i<this->size();i++){
					this->operator[](i)-=value;
			 }
			 return true;
		}
		return false;
	}
	
	template<class Type>
	bool ColumnBaseTyped<Type>::minus(ColumnPtr column){
		//std::transform ( first, first+5, second, results, std::plus<int>() );		
		shared_pointer_namespace::shared_ptr<ColumnBaseTyped<Type> > typed_column = shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<Type> >(column);
		if(!column) return false;
		for(unsigned int i=0;i<this->size();i++){
			this->operator[](i)-=typed_column->operator[](i);
		}			
		return true;
	}	


	template<class Type>
	bool ColumnBaseTyped<Type>::multiply(const boost::any& new_value){
		if(new_value.empty()) return false;
		if(typeid(Type)==new_value.type()){
			 Type value = boost::any_cast<Type>(new_value);
			 for(unsigned int i=0;i<this->size();i++){
					this->operator[](i)*=value;
			 }
			 return true;
		}
		return false;
	}
	
	template<class Type>
	bool ColumnBaseTyped<Type>::multiply(ColumnPtr column){
		//std::transform ( first, first+5, second, results, std::plus<int>() );		
		shared_pointer_namespace::shared_ptr<ColumnBaseTyped<Type> > typed_column = shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<Type> >(column);
		if(!column) return false;
		for(unsigned int i=0;i<this->size();i++){
			this->operator[](i)*=typed_column->operator[](i);
		}			
		return true;
	}



	template<class Type>
	bool ColumnBaseTyped<Type>::division(const boost::any& new_value){
		if(new_value.empty()) return false;
		if(typeid(Type)==new_value.type()){
			 Type value = boost::any_cast<Type>(new_value);
			 //check that we do not devide by zero
			 if(value==0) return false;
			 for(unsigned int i=0;i<this->size();i++){
					this->operator[](i)/=value;
			 }
			 return true;
		}
		return false;
	}
	
	template<class Type>
	bool ColumnBaseTyped<Type>::division(ColumnPtr column){
		//std::transform ( first, first+5, second, results, std::plus<int>() );		
		shared_pointer_namespace::shared_ptr<ColumnBaseTyped<Type> > typed_column = shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<Type> >(column);
		if(!column) return false;
		for(unsigned int i=0;i<this->size();i++){
                        //if(typed_column->operator[](i)==0) return false;
                        if(typed_column->operator[](i)==0){
                            std::cerr << "Fatal Error! In ColumnBaseTyped<Type>::division: Division by Zero!" << std::endl;
                        }
                        assert(typed_column->operator[](i)!=0);
			this->operator[](i)/=typed_column->operator[](i);
		}			
		return true;
	}

	//total tempalte specializations, because numeric computations are undefined on strings 
	template<>
	inline bool ColumnBaseTyped<std::string>::add(const boost::any&){ return false;	}
	template<>
	inline bool ColumnBaseTyped<std::string>::add(ColumnPtr){ return false;	}

	template<>
	inline bool ColumnBaseTyped<std::string>::minus(const boost::any&){ return false;	}
	template<>
	inline bool ColumnBaseTyped<std::string>::minus(ColumnPtr){ return false;	}


	template<>
	inline bool ColumnBaseTyped<std::string>::multiply(const boost::any&){ return false;	}
	template<>
	inline bool ColumnBaseTyped<std::string>::multiply(ColumnPtr){ return false;	}
	
	template<>
	inline bool ColumnBaseTyped<std::string>::division(const boost::any&){ return false;	}
	template<>
	inline bool ColumnBaseTyped<std::string>::division(ColumnPtr){ return false;	}

}; //end namespace CogaDB

