
#pragma once

#include <core/column_base_typed.hpp>
#include <core/column.hpp>
#include <core/base_table.hpp>
#include <lookup_table/lookup_column.hpp>

//#ifdef ENABLE_CDK_USAGE
//        #include <hardware_optimizations/primitives.hpp>
//#endif
//#include "runtime_configuration.hpp"

namespace CoGaDB{

/*!
 * 
 * 
 *  \brief     A LookupArray is a LookupColumn which is applied on a materialized column (of the table that is indexed by the Lookup column) and hence has a Type. 
 * 				This class represents a column with type T, which is essentially a tid list describing which values of a typed materialized column are included in the LookupArray.
 *  \details   This class is indentended to be a base class, so it has a virtual destruktor and pure virtual methods, which need to be implemented in a derived class. 
 *  \author    Sebastian Breß
 *  \version   0.2
 *  \date      2013
 *  \copyright GNU LESSER GENERAL PUBLIC LICENSE - Version 3, http://www.gnu.org/licenses/lgpl-3.0.txt
 */	

template<class T>
class LookupArray : public ColumnBaseTyped<T>{
	public:
	/***************** constructors and destructor *****************/
	LookupArray(const std::string& name, AttributeType db_type, ColumnPtr column, PositionListPtr tids);
	virtual ~LookupArray();

	virtual bool insert(const boost::any& new_Value);
        virtual bool insert(const T& new_Value);
	virtual bool update(TID tid, const boost::any& new_value);
	virtual bool update(PositionListPtr tid, const boost::any& new_value);	
	
	virtual bool remove(TID tid);
	//assumes tid list is sorted ascending
	virtual bool remove(PositionListPtr tid);
        virtual bool clearContent();
        
#ifdef ENABLE_CDK_USAGE
        virtual const PositionListPtr selection(const boost::any& value_for_comparison, const ValueComparator comp);  
        virtual const PositionListPtr selection(ColumnPtr comparison_column, const ValueComparator comp);
        virtual const PositionListPairPtr hash_join(ColumnPtr join_column_);
        virtual const PositionListPairPtr radix_join(ColumnPtr join_column_);
#endif
        
	virtual const boost::any get(TID tid);
	//virtual const boost::any* const getRawData()=0;
	virtual void print() const throw();
	virtual size_t size() const throw();
	virtual unsigned int getSizeinBytes() const throw();
        virtual PositionListPtr getPositionList();
        virtual shared_pointer_namespace::shared_ptr<ColumnBaseTyped<T> > getIndexedColumn();

	virtual const ColumnPtr copy() const;
        virtual const ColumnPtr gather(PositionListPtr tid_list);

	virtual bool store(const std::string& path);
	virtual bool load(const std::string& path);
	virtual bool isMaterialized() const  throw();
	virtual bool isCompressed() const  throw();	
	virtual const ColumnPtr materialize() throw();
        T* materializeToArray() throw();

	
	/*! \brief returns type information of internal values*/
	virtual T& operator[](const int index);
	//inline T& operator[](const int index) __attribute__((always_inline));

	private:
	shared_pointer_namespace::shared_ptr<ColumnBaseTyped<T> > column_;
	PositionListPtr tids_;
};

ColumnPtr createLookupArrayForColumn(ColumnPtr col, PositionListPtr tids_);
PositionListPtr getPositonListfromLookupArray(ColumnPtr col);


//inline ColumnPtr createLookupArrayForColumn(ColumnPtr col, PositionListPtr tids_)
//{
//    assert(col!=NULL);
//    
//    AttributeType type=col->getType();   
//    std::string column_name=col->getName();
//    
//    if(type==INT){
//            ColumnPtr lookup_array(new LookupArray<int>(column_name,INT,col,tids_));
//            return lookup_array;
//    }else if(type==FLOAT){
//            return ColumnPtr(new LookupArray<float>(column_name,FLOAT,col,tids_));
//    }else if(type==VARCHAR){
//            return ColumnPtr(new LookupArray<std::string>(column_name,VARCHAR,col,tids_));
//    }else if(type==BOOLEAN){
//            std::cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << column_name 
//                      << " Note: bool is currently not supported, will be added again in the future!"<< std::endl;
//            return ColumnPtr();
//    }else{
//            std::cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << column_name << std::endl;
//            return ColumnPtr();
//    }
//}
//
//inline PositionListPtr getPositonListfromLookupArray(ColumnPtr col){
//    if(!col) return PositionListPtr(); 
//    //is LookupArray?
//    if(col->isMaterialized() || col->isCompressed()){
//        //input is not LookupArray, so no Positionlist can be retrieved
//        return PositionListPtr(); 
//    }
//    
//    AttributeType type=col->getType();   
//
//    
//    if(type==INT){
//        shared_pointer_namespace::shared_ptr<LookupArray<int> > lookup_array = shared_pointer_namespace::dynamic_pointer_cast<LookupArray<int> > (col);
//        return lookup_array->getPositionList();           
//    }else if(type==FLOAT){
//        shared_pointer_namespace::shared_ptr<LookupArray<float> > lookup_array = shared_pointer_namespace::dynamic_pointer_cast<LookupArray<float> > (col);
//        return lookup_array->getPositionList();              
//    }else if(type==VARCHAR){
//        shared_pointer_namespace::shared_ptr<LookupArray<std::string> > lookup_array = shared_pointer_namespace::dynamic_pointer_cast<LookupArray<std::string> > (col);
//        return lookup_array->getPositionList();               
//    }else if(type==BOOLEAN){
//            std::cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << col->getName() 
//                      << " Note: bool is currently not supported, will be added again in the future!"<< std::endl;
//            return PositionListPtr();
//    }else{
//            std::cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " <<  col->getName() << std::endl;
//            return PositionListPtr();
//    }
//    
//
//}
//
//
//	//typedef shared_pointer_namespace::shared_ptr<LookupArray> LookupArrayPtr;
//
///***************** Start of Implementation Section ******************/
//
//	
//	template<class T>
//	LookupArray<T>::LookupArray(const std::string& name, AttributeType db_type, ColumnPtr column, PositionListPtr tids) 
//						: ColumnBaseTyped<T>(name, db_type),
//						  column_( shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<T> > (column) ),
//						  tids_(tids) {
//	
//		assert(column_!=NULL);
//		assert(tids_!=NULL);
//		assert(db_type==column->getType());
//                this->fk_constr_=column->getForeignKeyConstraint();
//                this->has_foreign_key_constraint_=column->hasForeignKeyConstraint();
//	}
//
//	template<class T>
//	LookupArray<T>::~LookupArray(){
//
//	}
//
//	template<class T>
//	bool LookupArray<T>::insert(const boost::any& ){
//		return false;
//	}
//	template<class T>
//	bool LookupArray<T>::insert(const T& new_Value){
//		return false;
//	} 
//	template<class T>
//	bool LookupArray<T>::update(TID , const boost::any& ){
//		return  false;
//	}
//
//	template<class T>
//	bool LookupArray<T>::update(PositionListPtr, const boost::any&){
//		return  false;			
//	}
//	
//	template<class T>
//	bool LookupArray<T>::remove(TID ){
//
//		return false;
//	}
//	
//	//assumes tid list is sorted ascending
//	template<class T>
//	bool LookupArray<T>::remove(PositionListPtr){
//		return false;
//	}
//        
//	template<class T>
//	bool LookupArray<T>::clearContent(){
//            return false;
//        }
//        
//        
//#ifdef ENABLE_CDK_USAGE
//        template<class T>
//        const PositionListPtr LookupArray<T>::selection(const boost::any& value_for_comparison, const ValueComparator comp){
//            
////            T value;
////            if(value_for_comparison.type()!=typeid(T)){ 
////                //catch some special cases
////                if(typeid(T)==typeid(float) && value_for_comparison.type()==typeid(int)){
////                    value = boost::any_cast<int>(value_for_comparison);
////                }else{
////                    std::cout << "Fatal Error!!! Typemismatch for column " << this->name_ << std::endl;
////                    std::cout << "Column Type: " << typeid(T).name() << " filter value type: " << value_for_comparison.type().name() << std::endl;
////                    std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
////                    exit(-1);
////                }
////            }else{
////                //everything fine, filter value matches type of column
////                value = boost::any_cast<T>(value_for_comparison);
////            }  
//            
//            if( (this->column_->size()==0) || (this->tids_->size()==0)){
//                return PositionListPtr(new PositionList());
//            }
//            
//            if(this->column_->isMaterialized() && !this->column_->isCompressed()){
//                shared_pointer_namespace::shared_ptr<Column<T> > base_column = shared_pointer_namespace::dynamic_pointer_cast<Column<T> >(this->column_);
//                assert(base_column!=NULL);
//            
//                T* base_array = hype::util::begin_ptr(base_column->getContent());
//                assert(base_array!=NULL);
//
//                TID* tids = this->tids_->data();
//                assert(tids!=NULL);
//                
//                T* input_column = (T*) malloc(sizeof(T)*this->tids_->size());
//                //CoGaDB::RuntimeConfiguration::instance().
//                //fetch the relevant tids specified in TID list
//                //for very small arrays, use serial gather to reduce threading overhead
//                if(this->tids_->size()<300){
//                        CDK::util::serial_gather(base_array, tids, this->tids_->size(), input_column);
//                }else{
//                        CDK::util::parallel_gather(base_array, tids, this->tids_->size(), input_column, boost::thread::hardware_concurrency());
//                }
//                
//                //perform parallel selection on the generated input column
//                PositionListPtr result = CDK::selection::parallel_selection(input_column, this->tids_->size(), value_for_comparison, comp, boost::thread::hardware_concurrency());
//                free(input_column);
//                return result;
//            }else{
//                //for strings we can use the default method
//                return CoGaDB::ColumnBaseTyped<T>::selection(value_for_comparison, comp);
//            }
//        }
//        
//        template<class T>
//        const PositionListPtr LookupArray<T>::selection(ColumnPtr comparison_column, const ValueComparator comp){
//            //same behaviour as in parent class
//            return CoGaDB::ColumnBaseTyped<T>::selection(comparison_column, comp);
//        }
//        
//        template<class T>
//	const PositionListPairPtr LookupArray<T>::hash_join(ColumnPtr join_column_){
//            assert(this->size()<=join_column_->size()); 
//            if(join_column_->type()!=typeid(T)){
//                    std::cout << "Fatal Error!!! Typemismatch for columns " << this->name_  << " and " << join_column_->getName() << std::endl;
//                    std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
//                    exit(-1);
//            }
//            T* this_array=this->materializeToArray();
//            assert(this_array!=NULL);
//            T* input_column=NULL;
//            PositionListPairPtr ret;
//            if(join_column_->isMaterialized()){
//                shared_pointer_namespace::shared_ptr<Column<T> > join_column = shared_pointer_namespace::dynamic_pointer_cast<Column<T> >(join_column_);
//                assert(join_column!=NULL);
//                input_column=join_column->data();
//                assert(input_column!=NULL);
//        
//                ret = CDK::join::serial_hash_join(this_array, this->size(), input_column, join_column->size());
//
//
//            }else if (!join_column_->isMaterialized() && !join_column_->isCompressed()){ //Lookup Array?
//                shared_pointer_namespace::shared_ptr<LookupArray<T> > join_column = shared_pointer_namespace::dynamic_pointer_cast<LookupArray<T> >(join_column_);
//                assert(join_column!=NULL); 
//
//                T* input_column = join_column->materializeToArray();
//
//                assert(this_array!=NULL);        
//                ret = CDK::join::serial_hash_join(this_array, this->size(), input_column, join_column->size());
//                //ok, job done, cleanup temporary array
//                free(input_column);    
//            }else if(join_column_->isCompressed()){
//                //if compressed, use old style hash join via ColumnBaseTyped class
//                ret = CoGaDB::ColumnBaseTyped<T>::hash_join(join_column_);
//            }else{
//                //detected impossible case, abort
//                COGADB_FATAL_ERROR("Detected Invalid Column Class! Not MAterialized, No LookupARray and no CompressedColumn!","");
//                ret = PositionListPairPtr();
//            }
//            free(this_array);
//            return ret;
//        }    
//        
//        
//        template<class T>
//	const PositionListPairPtr LookupArray<T>::radix_join(ColumnPtr join_column_){
//            COGADB_FATAL_ERROR("called non INT instantiation of LookupArray<T>::radix_join, which only supports integer columns!","");
//            return PositionListPairPtr(); 
//        }
//        
//        template<>
//	inline const PositionListPairPtr LookupArray<int>::radix_join(ColumnPtr join_column_){
//            assert(this->size()<=join_column_->size());
//            if(join_column_->type()!=typeid(int)){
//                    std::cout << "Fatal Error!!! Typemismatch for columns " << this->name_  << " and " << join_column_->getName() << std::endl;
//                    std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
//                    exit(-1);
//            }
//            int* this_array=this->materializeToArray();
//            assert(this_array!=NULL);
//            int* input_column=NULL;
//            PositionListPairPtr ret;
//            if(join_column_->isMaterialized()){
//                shared_pointer_namespace::shared_ptr<Column<int> > join_column = shared_pointer_namespace::dynamic_pointer_cast<Column<int> >(join_column_);
//                assert(join_column!=NULL);
//                input_column=join_column->data();
//                assert(input_column!=NULL);
//        
//                ret = CDK::join::radix_join(this_array, this->size(), input_column, join_column->size());
//
//
//            }else if (!join_column_->isMaterialized() && !join_column_->isCompressed()){ //Lookup Array?
//                shared_pointer_namespace::shared_ptr<LookupArray<int> > join_column = shared_pointer_namespace::dynamic_pointer_cast<LookupArray<int> >(join_column_);
//                assert(join_column!=NULL); 
//
//                int* input_column = join_column->materializeToArray();
//
//                assert(this_array!=NULL);        
//                ret = CDK::join::radix_join(this_array, this->size(), input_column, join_column->size());
//                //ok, job done, cleanup temporary array
//                free(input_column);    
//            }else if(join_column_->isCompressed()){
//                //if compressed, use old style hash join via ColumnBaseTyped class
//                ret = CoGaDB::ColumnBaseTyped<int>::hash_join(join_column_);
//            }else{
//                //detected impossible case, abort
//                COGADB_FATAL_ERROR("Detected Invalid Column Class! Not MAterialized, No LookupARray and no CompressedColumn!","");
//                ret = PositionListPairPtr();
//            }
//            free(this_array);
//            return ret;
//        }
//        
//#endif
//        
//	
//	template<class T>
//	const boost::any LookupArray<T>::get(TID tid){
//		return boost::any((*this)[tid]);
//	}
//
//	template<class T>
//	void LookupArray<T>::print() const throw(){
//		
//		const shared_pointer_namespace::shared_ptr<ColumnBaseTyped<T> > column = column_;
//		const PositionListPtr tids = tids_;
//
//		std::cout << "Lookup Array for Column " << column_->getName() << " ";
//		if(column_->isMaterialized()){
//			std::cout << "which is a materialized Column" << std::endl;
//		}else{
//			std::cout << "which is a LookupArray of a Lookup column" << std::endl;
//		}
//		std::cout << "| values | Translatetion TIDS | Index in Lookup Table |" << std::endl;
//		for(unsigned int i=0;i<tids->size();i++){
//			std::cout << "| " << (*column_)[(*tids_)[i]] << " | " << (*tids_)[i] << " | " << i << " |" << std::endl;
//		}
//
//
//	}
//	template<class T>
//	size_t LookupArray<T>::size() const throw(){
//
//		return tids_->size();
//	}
//
//	template<class T>
//	const ColumnPtr LookupArray<T>::materialize() throw(){
//
//            Column<T>* result = new Column<T>(this->name_,this->db_type_);
//            std::vector<T>& data = result->getContent();
//            data.resize(this->size());
//            for(unsigned int i=0;i<this->size();i++){
//		data[i]=(*this)[i];
//            }
//            return ColumnPtr(result); 
//        }
//        
//	template<class T>
//        T* LookupArray<T>::materializeToArray() throw(){
//                if(this->size()==0 || this->column_->size()==0) return NULL;
//                shared_pointer_namespace::shared_ptr<Column<T> > base_column = shared_pointer_namespace::dynamic_pointer_cast<Column<T> >(this->column_);
//                assert(base_column!=NULL);
//                T* base_column_array = base_column->data(); //hype::util::begin_ptr(base_column->values_);
//                assert(base_column_array!=NULL);
//                //allocate temporary array
//                T* input_column = (T*) malloc(sizeof(T)*this->size());
//                assert(input_column!=NULL);
//                //CoGaDB::RuntimeConfiguration::instance().
//                //fetch the relevant tids specified in TID list
//                //for very small arrays, use serial gather to reduce threading overhead
////                CDK::util::serial_gather(base_column_array, tids->data(), tids->size(), input_column);
//                if(this->size()<300){
//                        CDK::util::serial_gather(base_column_array, this->tids_->data(), this->size(), input_column);
//                }else{
//                        CDK::util::parallel_gather(base_column_array, this->tids_->data(), this->size(), input_column, boost::thread::hardware_concurrency());
//                }
//                return input_column;
//        }
//        
//
//	template<class T>
//	const ColumnPtr LookupArray<T>::copy() const{
//		PositionListPtr new_tids (new PositionList(*tids_));
//		return ColumnPtr(new LookupArray<T>(this->name_,this->db_type_,this->column_,new_tids));
//	}
//        
//        template<class T>
//        const ColumnPtr LookupArray<T>::gather(PositionListPtr tid_list){
//            Column<T>* result = new Column<T>(this->name_,this->db_type_);
//            std::vector<T>& data = result->getContent();
//            data.resize(tid_list->size());
//            for(unsigned int i=0;i<tid_list->size();i++){              
//		data[i]=(*this)[(*tid_list)[i]];
//            }
//            return ColumnPtr(result); 
//        }    
//        
//	/***************** relational operations on LookupArrays which return lookup tables *****************/
////	template<class T>
////	const std::vector<TID> LookupArray<T>::sort(const ComputeDevice comp_dev) const {
//
////		return std::vector<TID>();
////	}
//
////	template<class T> 
////	const std::vector<TID> LookupArray<T>::selection(const boost::any& value_for_comparison, const ValueComparator comp, const ComputeDevice comp_dev) const {
////		return std::vector<TID>();
////	}
////	//join algorithms
////	template<class T>
////	const std::vector<TID_Pair> LookupArray<T>::sort_merge_join(ColumnPtr join_LookupArray, const ComputeDevice comp_dev) const{
////		return std::vector<TID_Pair>();
////	}
////	template<class T>
////	const std::vector<TID_Pair> LookupArray<T>::nested_loop_join(ColumnPtr join_LookupArray, const ComputeDevice comp_dev) const{
////		return false;
////	}
//	template<class T>
//	bool LookupArray<T>::store(const std::string&){
//		return false;
//	}
//	template<class T>
//	bool LookupArray<T>::load(const std::string&){
//		return false;
//	}
//	template<class T>
//	bool LookupArray<T>::isMaterialized() const  throw(){
//		return false;
//	}
//	template<class T>
//	bool LookupArray<T>::isCompressed() const  throw(){
//		return false;
//	}
//	template<class T>
//	T& LookupArray<T>::operator[](const int index){
//		return (*column_)[(*tids_)[index]];
//	}
//
//	template<class T>
//	unsigned int LookupArray<T>::getSizeinBytes() const throw(){
//		return tids_->capacity()*sizeof(typename PositionList::value_type);
//	}
//
//        template<class T>
//	PositionListPtr LookupArray<T>::getPositionList(){
//            return this->tids_;
//        }
//        
//        template<class T>
//        shared_pointer_namespace::shared_ptr<ColumnBaseTyped<T> > LookupArray<T>::getIndexedColumn(){
//            return this->column_;
//        }
        
/***************** End of Implementation Section ******************/


}; //end namespace CogaDB

