
#include <core/gpu_column_cache.hpp>
#include <boost/thread.hpp>
#include <core/data_dictionary.hpp>

namespace CoGaDB{
using namespace std;
		boost::mutex global_gpu_cache_mutex;

		GPU_Column_Cache::GPU_Column_Cache(unsigned int max_gpu_buffer_size_in_byte) : 
                	map_(),
                        join_index_map_(),
			max_gpu_buffer_size_in_byte_(max_gpu_buffer_size_in_byte), 
			caching_enabled_(true)
                {

		}

		GPU_Column_Cache& GPU_Column_Cache::instance(){
			static GPU_Column_Cache cache(getTotalGPUMemorySizeInByte()*0.5); //use half of the total device memory as buffer for columns 
			return cache;
		}	

		void GPU_Column_Cache::printStatus() const throw(){
			boost::lock_guard<boost::mutex> lock(global_gpu_cache_mutex);
			cout << "GPU_Column_Cache Status:" << endl;
			cout << "GPU Buffer Size: " << max_gpu_buffer_size_in_byte_/(1024*1024) << "MB" << endl; 
			cout << "Total GPU Memory: " << getTotalGPUMemorySizeInByte()/(1024*1024) << "MB" << endl; 
			cout << "Free GPU Memory: " << getFreeGPUMemorySizeInByte()/(1024*1024) << "MB" << endl; 
			cout << "Cached columns:" << endl;
			Map::const_iterator cit;
			for(cit=map_.begin();cit!=map_.end();++cit){
				cout << cit->second->getName() << " of Size " << cit->second->size() << " elements, currently in use: " << !cit->second.unique() << endl;
			}
		}

                bool isPersistentColumn(ColumnPtr column_ptr){
                    
                    std::list<std::pair<ColumnPtr,TablePtr> > columns = DataDictionary::instance().getColumnsforColumnName(column_ptr->getName());
                    //we assume unique column names
                    assert(columns.size()<=1);
                    if(columns.size()==1){
                        if(columns.front().first==column_ptr){
                            return true;
                        }else{
                            return false;
                        }
                    }else{
                        return false;
                    }
                }
                
                bool isIntermediateResultColumn(ColumnPtr column_ptr){
                    return !isPersistentColumn(column_ptr);
                }
                
		const gpu::GPU_Base_ColumnPtr GPU_Column_Cache::getGPUColumn(ColumnPtr column_ptr){
			boost::lock_guard<boost::mutex> lock(global_gpu_cache_mutex);
                        
			gpu::GPU_Base_ColumnPtr result_ptr;
			if(!column_ptr) return gpu::GPU_Base_ColumnPtr();
                        //do not cache if caching is disabled or column is intermediate result
			if(!caching_enabled_ || isIntermediateResultColumn(column_ptr)){
				result_ptr=gpu::copy_column_host_to_device(column_ptr);
				return result_ptr;
			}

			Map::iterator it=map_.find(column_ptr);
			if(it==map_.end()){
				cout << "Cache Miss: " << column_ptr->getName() << endl;

				unsigned int used_memory_in_bytes=getTotalGPUMemorySizeInByte()-getFreeGPUMemorySizeInByte();

				if(column_ptr->getSizeinBytes()+used_memory_in_bytes>max_gpu_buffer_size_in_byte_){
					cout << "Warning: GPU_Column_Cache::getGPUColumn(): Unsufficient Device Memory" << endl;
					//apply column replacement strategy!
					return gpu::GPU_Base_ColumnPtr();
				}

				result_ptr=gpu::copy_column_host_to_device(column_ptr);
				if(!result_ptr){ 
					cout << "Warning: GPU_Column_Cache::getGPUColumn(): Unsufficient Device Memory" << endl;
					return gpu::GPU_Base_ColumnPtr();
				}		
				//if(caching_enabled_){
					map_.insert( std::make_pair(column_ptr, result_ptr));
				//}
				return result_ptr;
			}else{
				cout << "Cache Hit: " << column_ptr->getName() << endl;
				return it->second; //return pointer to GPU memory
			}

		}

                const gpu::GPU_JoinIndexPtr GPU_Column_Cache::getGPUJoinIndex(JoinIndexPtr join_index){
			boost::lock_guard<boost::mutex> lock(global_gpu_cache_mutex); 
                    
			gpu::GPU_JoinIndexPtr result_ptr;
			if(!join_index) return gpu::GPU_JoinIndexPtr();
			if(!caching_enabled_){
				result_ptr=gpu::copy_JoinIndex_host_to_device(join_index);
				return result_ptr;
			}

			JoinIndexMap::iterator it=join_index_map_.find(join_index);
			if(it==join_index_map_.end()){
				cout << "Cache Miss: JOIN Index"  << endl;

				unsigned int used_memory_in_bytes=getTotalGPUMemorySizeInByte()-getFreeGPUMemorySizeInByte();

                                size_t size_of_joinindex=(join_index->first->getPositionList()->size()+join_index->second->getPositionList()->size())*sizeof(PositionList::value_type);
				if(size_of_joinindex+used_memory_in_bytes>max_gpu_buffer_size_in_byte_){
					COGADB_WARNING("GPU_Column_Cache::getGPUJoinIndex(): Unsufficient Device Memory","");
					//apply column replacement strategy!
					return gpu::GPU_JoinIndexPtr();
				}

				result_ptr=gpu::copy_JoinIndex_host_to_device(join_index);
				if(!result_ptr){ 
					COGADB_WARNING("GPU_Column_Cache::getGPUJoinIndex(): Unsufficient Device Memory","");
					return gpu::GPU_JoinIndexPtr();
				}		
				//if(caching_enabled_){
					join_index_map_.insert( std::make_pair(join_index, result_ptr));
				//}
				return result_ptr;
			}else{
				cout << "Cache Hit: JOIN INDEX" << endl;
				return it->second; //return pointer to GPU memory
			}

                    return gpu::GPU_JoinIndexPtr();
                }

		bool GPU_Column_Cache::removeGPUColumn(gpu::GPU_Base_ColumnPtr dev_col){
                    	boost::lock_guard<boost::mutex> lock(global_gpu_cache_mutex);
			Map::iterator it;
			for(it=map_.begin();it!=map_.end();++it){
				if(it->second==dev_col){
					map_.erase(it);
					return true;
				}
			}
			return true;
		}
                
                bool GPU_Column_Cache::removeGPUJoinIndex(gpu::GPU_JoinIndexPtr join_index){
                    boost::lock_guard<boost::mutex> lock(global_gpu_cache_mutex);
			JoinIndexMap::iterator it;
			for(it=join_index_map_.begin();it!=join_index_map_.end();++it){
				if(it->second==join_index){
					join_index_map_.erase(it);
					return true;
				}
			}
                    return false;
                }


}; //end namespace CoGaDB
