
//#include <gpu/gpu_bitmap.hpp>
#include <gpu/gpu_base_column.hpp>
#include <gpu/gpu_algorithms.hpp>
#include <gpu/gpu_positionlist.hpp>
#include <thrust/device_vector.h>
#include <thrust/scan.h>


#define gpuErrchk(ans) { gpuAssert((ans), __FILE__, __LINE__); }
inline void gpuAssert(cudaError_t code, char *file, int line, bool abort=true)
{
   if (code != cudaSuccess) 
   {
      fprintf(stderr,"GPUassert: %s %s %d\n", cudaGetErrorString(code), file, line);
      if (abort) exit(code);
   }
}

namespace CoGaDB{
        
    namespace gpu{
        
    GPU_Bitmap::GPU_Bitmap(char* bitmap, size_t num_of_bits) : bitmap_(bitmap), num_of_bits_(num_of_bits){
        assert(bitmap_!=NULL);
    }
    GPU_Bitmap::GPU_Bitmap(size_t num_of_bits, bool init_value, bool initialize) : bitmap_(NULL), num_of_bits_(num_of_bits){
        cudaError_t err = cudaMalloc((void**) &bitmap_,(num_of_bits+7)/8);
        assert(bitmap_!=NULL);
        if(initialize){
            if(init_value){
                err = cudaMemset(bitmap_, 255, (num_of_bits+7)/8);
            }else{
                err = cudaMemset(bitmap_, 0, (num_of_bits+7)/8);
            }
	
        }

    }
    GPU_Bitmap::~GPU_Bitmap(){
        if(bitmap_){
            cudaError_t err = cudaFree(bitmap_);
            bitmap_=NULL;
        }
    }
    char* GPU_Bitmap::data(){
        return bitmap_;
    }
    size_t GPU_Bitmap::size() const{
        return num_of_bits_;
    }
    
    const BitmapPtr copy_Bitmap_device_to_host(GPU_BitmapPtr gpu_bitmap){
        BitmapPtr bitmap(new Bitmap(gpu_bitmap->size(),false,false));
        char* cpu_bitmap = bitmap->data();
//        char* gpu_bitmap = gpu_bitmap->data();
        cudaError_t err = cudaMemcpy(cpu_bitmap, gpu_bitmap->data(), sizeof(char)*((gpu_bitmap->size()+7)/8), cudaMemcpyDeviceToHost);
        gpuErrchk( cudaPeekAtLastError() );
        gpuErrchk( cudaDeviceSynchronize() );
        return bitmap;
    }
    const GPU_BitmapPtr copy_Bitmap_host_to_device(BitmapPtr bitmap){
        GPU_BitmapPtr gpu_bitmap(new GPU_Bitmap(bitmap->size(),false,false));
        char* plain_gpu_bitmap = gpu_bitmap->data();
//        char* gpu_bitmap = gpu_bitmap->data();
        cudaError_t err = cudaMemcpy(plain_gpu_bitmap, bitmap->data(), sizeof(char)*((bitmap->size()+7)/8), cudaMemcpyHostToDevice);
        gpuErrchk( cudaPeekAtLastError() );
        gpuErrchk( cudaDeviceSynchronize() );
        return gpu_bitmap;
    }
    
    


////Note: assume result_bitmap is set to zero! (use cudaMemset)
//__global__ void convertPositionListToBitmap_kernel(TID* tids, size_t num_tids, size_t num_rows_base_table, unsigned int* result_bitmap){
//
//    int tid = ...
//    while(...){
//    unsigned int bitmap_tid = tids[i] / 32; 
//    unsigned int bitmap_offset = tids[i] % 32; // %8
//    
//    char bitmask = 1 << bitmap_offset;
//
//    /*
//    char val = atomic_read(bitmap[i]);
//    val |= bitmask;
//    atomic_write(bitmask[i],val);
//    */    
//    atomicOr(&bitmap[bitmap_tid], bitmask);
//    }
//}

//set memory to 0 beforehand!
__global__ void convertPositionListToBitmap_create_flagarray_kernel(TID* tids, size_t num_tids, size_t num_rows_base_table, char* flag_array){

    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    while(tid<num_tids){
        flag_array[tids[tid]]=1;
        tid+=blockDim.x * gridDim.x;
    }
        
}

//GPU_BitmapPtr convertPositionListToBitmap(GPU_Positionlist gpu_tids);
//
//GPU_Positionlist convertBitmapToPositionList(GPU_BitmapPtr gpu_bitmap);
//


__global__ void convertPositionListToBitmap_pack_flagarray_kernel(char* flag_array, size_t num_rows_base_table, char* result_bitmap){

    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    while(tid<(num_rows_base_table+7)/8){
    int flag_array_tid = tid*8;
    char packed_bits=0;
    packed_bits |= (flag_array[flag_array_tid]!=0);
    packed_bits |= (flag_array[flag_array_tid+1]!=0) << 1;
    packed_bits |= (flag_array[flag_array_tid+2]!=0) << 2;
    packed_bits |= (flag_array[flag_array_tid+3]!=0) << 3;
    packed_bits |= (flag_array[flag_array_tid+4]!=0) << 4;
    packed_bits |= (flag_array[flag_array_tid+5]!=0) << 5;
    packed_bits |= (flag_array[flag_array_tid+6]!=0) << 6;
    packed_bits |= (flag_array[flag_array_tid+7]!=0) << 7;
        
    result_bitmap[tid]=packed_bits;
    tid+=blockDim.x * gridDim.x;
    }
}

__global__ void convertBitmapToPositionList_setFlagArray_kernel(char* bitmap, size_t num_of_bits, char* flag_array){
    int tid = (threadIdx.x + blockIdx.x * blockDim.x)*8;
    int flag_array_size = (num_of_bits+7)/8;
    flag_array_size *=8;
    while(tid<flag_array_size){//num_of_bits-8){
       flag_array[tid]=((bitmap[tid/8] & 1)!=0);
       flag_array[tid+1]=((bitmap[tid/8] & 2)!=0);
       flag_array[tid+2]=((bitmap[tid/8] & 4)!=0);
       flag_array[tid+3]=((bitmap[tid/8] & 8)!=0);
       flag_array[tid+4]=((bitmap[tid/8] & 16)!=0);
       flag_array[tid+5]=((bitmap[tid/8] & 32)!=0);
       flag_array[tid+6]=((bitmap[tid/8] & 64)!=0);
       flag_array[tid+7]=((bitmap[tid/8] & 128)!=0);
       tid+=(blockDim.x * gridDim.x)*8;
    }
//    while(tid<flag_array_size){//(num_of_bits+7)/8){
//       for(unsigned int i=0;i<num_of_bits%8;++i){ //process part of last byte in bitmap
//          flag_array[tid+i]=bitmap[(tid+i)/8] & (1 << i);
//       }
//       //tid+=blockDim.x * gridDim.x;
//       tid+=(blockDim.x * gridDim.x)*8;
//    }
    
}

__global__ void convertBitmapToPositionList_createPositionList_kernel(char* flag_array, size_t flag_array_size, unsigned int* write_positions, TID* result_tids){

  int tid = threadIdx.x + blockIdx.x * blockDim.x;
  while(tid<flag_array_size){
  if(flag_array[tid]){ 
    unsigned int write_index = write_positions[tid];
    result_tids[write_index] = tid;
  }
    tid+=blockDim.x * gridDim.x;
  }
}

__global__ void bitmap_and_kernel(char* left_bitmap, char* right_bitmap, size_t num_of_bits, char* result_bitmap){

  int tid = threadIdx.x + blockIdx.x * blockDim.x;
  while(tid<(num_of_bits+7)/8){  
    result_bitmap[tid] = left_bitmap[tid] & right_bitmap[tid];
    tid+=blockDim.x * gridDim.x;
  }
}

const GPU_BitmapPtr GPU_Operators::convertPositionListToBitmap(GPU_PositionlistPtr gpu_pos_list, size_t num_rows_base_table){
    assert(gpu_pos_list!=NULL);
    TID* gpu_tids = getRawPointerToGPU_Positionlist(gpu_pos_list);
    assert(gpu_tids!=NULL);
    
    const int number_of_blocks=512;
    const int number_of_threads_per_block=1024;

//    convertPositionListToBitmap_create_flagarray_kernel(TID* tids, size_t num_tids, size_t num_rows_base_table, char* flag_array);
//    convertPositionListToBitmap_pack_flagarray_kernel(char* flag_array, size_t num_rows_base_table, char* result_bitmap);
    unsigned int number_of_bytes_for_flag_array= (num_rows_base_table+7)/8;
    number_of_bytes_for_flag_array *=8;
    thrust::device_ptr<char> flag_array = thrust::device_malloc<char>(number_of_bytes_for_flag_array);//num_rows_base_table);
    assert(flag_array.get()!=NULL);
    cudaError_t err = cudaMemset(flag_array.get(), 0, num_rows_base_table);
    gpuErrchk( err );

    convertPositionListToBitmap_create_flagarray_kernel<<<number_of_blocks,number_of_threads_per_block>>>(gpu_tids, gpu_pos_list->size(), num_rows_base_table, flag_array.get());
    gpuErrchk( cudaPeekAtLastError() );
    gpuErrchk( cudaDeviceSynchronize() );    

    //thrust::exclusive_scan(flag_array, flag_array+input_array_size, write_positions_array);
    //create gpu bitmap and init bitmap with 0 (false)
    GPU_BitmapPtr gpu_bitmap(new GPU_Bitmap(num_rows_base_table,false,true));
    convertPositionListToBitmap_pack_flagarray_kernel<<<number_of_blocks,number_of_threads_per_block>>>(flag_array.get(), num_rows_base_table, gpu_bitmap->data());
    gpuErrchk( cudaPeekAtLastError() );
    gpuErrchk( cudaDeviceSynchronize() );
    
    thrust::device_free(flag_array);
                       
    return gpu_bitmap;
}

const GPU_PositionlistPtr GPU_Operators::convertBitmapToPositionList(GPU_BitmapPtr gpu_bitmap){
                       const int number_of_blocks=512;
                       const int number_of_threads_per_block=1024;
    
                       unsigned int input_array_size = gpu_bitmap->size();
                       unsigned int flag_array_size = (gpu_bitmap->size()+7)/8;
                       flag_array_size *= 8;
    
                       thrust::device_ptr<char> flag_array = thrust::device_malloc<char>(flag_array_size);
                       assert(flag_array.get()!=NULL);
                       cudaError_t err = cudaMemset(flag_array.get(), 0, flag_array_size);
                       gpuErrchk( err );
                       
                       std::cout << "flag array: " << (void*) flag_array.get() << std::endl;
                       thrust::device_ptr<TID> write_positions_array = thrust::device_malloc<TID>(input_array_size);
                       assert(write_positions_array.get()!=NULL);
                       

                       convertBitmapToPositionList_setFlagArray_kernel<<<number_of_blocks,number_of_threads_per_block>>>(gpu_bitmap->data(), gpu_bitmap->size(), flag_array.get());
                       gpuErrchk( cudaPeekAtLastError() );
                       gpuErrchk( cudaDeviceSynchronize() );

                       thrust::exclusive_scan(flag_array, flag_array+input_array_size, write_positions_array);
              
                       unsigned int result_size; // = write_positions_array[input_array_size-1];
                       char last_flag;

                       unsigned int * tmp =thrust::raw_pointer_cast(write_positions_array);
                       err = cudaMemcpy(&result_size, (void*)&tmp[input_array_size-1], sizeof(unsigned int), cudaMemcpyDeviceToHost);
//                       gpuErrchk(err);
                       gpuErrchk( cudaPeekAtLastError() );
                       gpuErrchk( cudaDeviceSynchronize() )

                       char* tmp2 = thrust::raw_pointer_cast(flag_array);
                       err = cudaMemcpy(&last_flag, (void*)&tmp2[input_array_size-1], sizeof(char), cudaMemcpyDeviceToHost);
//                       gpuErrchk(err);
                       gpuErrchk( cudaPeekAtLastError() );
                       gpuErrchk( cudaDeviceSynchronize() );

                       //since we perform an exclusive scan, to get the write positions, 
                       //we have to check whether the last element in the flag array was a 
                       //match and adjust the result size accordingly
                       if(last_flag) result_size++;
                       std::cout << "Bitmap to PositionList: #tids: " << result_size << std::endl;
    

    
                       thrust::device_vector<unsigned int> result_tids(result_size); //reserve space for result
                       assert(result_tids.size()>0);

                       convertBitmapToPositionList_createPositionList_kernel<<<number_of_blocks,number_of_threads_per_block>>>(flag_array.get(), input_array_size, write_positions_array.get(), thrust::raw_pointer_cast(&result_tids[0]));
                       gpuErrchk( cudaPeekAtLastError() );
                       gpuErrchk( cudaDeviceSynchronize() );
                       // deallocate with device_free
                       thrust::device_free(write_positions_array);
                       thrust::device_free(flag_array);

                       GPU_PositionlistPtr result(new Impl_GPU_Positionlist(result_tids));
                       return result;
  
}


const GPU_BitmapPtr GPU_Operators::computeBitmapOperationAnd(GPU_BitmapPtr left_bitmap, GPU_BitmapPtr right_bitmap){
    assert(left_bitmap!=NULL);
    assert(right_bitmap!=NULL);
    assert(left_bitmap->size()==right_bitmap->size());
    
    const int number_of_blocks=512;
    const int number_of_threads_per_block=512;
    size_t num_of_bits=left_bitmap->size();

    GPU_BitmapPtr gpu_bitmap(new GPU_Bitmap(num_of_bits,false,false));
    bitmap_and_kernel<<<number_of_blocks,number_of_threads_per_block>>>(left_bitmap->data(), right_bitmap->data(), num_of_bits, gpu_bitmap->data());
    return gpu_bitmap;
}

    
    };
}; //end namespace CoGaDB


