

#include <lookup_table/join_index.hpp>
#include <core/table.hpp>
#include <sstream>
#include <boost/thread.hpp>

namespace CoGaDB {
        boost::mutex global_join_index_mutex;
                        
        JoinIndexes& JoinIndexes::instance(){
            static JoinIndexes join_indx;
            return join_indx;
        }
    
        std::string JoinIndexes::joinExpressionToString(TablePtr pk_tab, std::string pk_tab_col_name, TablePtr fk_tab, std::string fk_tab_col_name){
            std::stringstream ss;
            ss << "JOIN_INDEX(" <<  pk_tab->getName() << "." << pk_tab_col_name << "," << fk_tab->getName() << "." << fk_tab_col_name << ")";
            return ss.str();
        }
        
        JoinIndexPtr JoinIndexes::getJoinIndex(TablePtr pk_tab, std::string pk_tab_col_name, TablePtr fk_tab, std::string fk_tab_col_name){
            boost::lock_guard<boost::mutex> lock(global_join_index_mutex);
//            assert(pk_tab->hasPrimaryKeyConstraint(pk_tab_col_name)==true);
//            assert(fk_tab->hasForeignKeyConstraint(fk_tab_col_name)==true);
            std::string index_name = joinExpressionToString(pk_tab, pk_tab_col_name, fk_tab, fk_tab_col_name);
            Indexes::iterator it = indeces_.find(index_name);
            if(it!=indeces_.end()){
                return it->second;
            }else{
                JoinIndexPtr join_index = createJoinIndex(pk_tab, pk_tab_col_name, fk_tab, fk_tab_col_name);
                this->indeces_.insert(std::make_pair(index_name,join_index));
                return join_index;
            }
            
        }
    
        TablePtr JoinIndexes::getSystemTable(){
            TableSchema result_schema;
            result_schema.push_back(Attribut(VARCHAR,"JOIN INDEX NAME"));
            result_schema.push_back(Attribut(VARCHAR,"PRIMARY KEY TABLE NAME"));
            result_schema.push_back(Attribut(VARCHAR,"FOREIGN KEY TABLE NAME"));

            TablePtr result_tab(new Table("SYS_JOIN_INDEXES",result_schema));
            Indexes::iterator it;
            for(it= indeces_.begin();it!=indeces_.end();++it){
                Tuple t;
                t.push_back(it->first);
                t.push_back(it->second->first->getTable()->getName());
                t.push_back(it->second->second->getTable()->getName());
                result_tab->insert(t);
            }
            return result_tab;
        }
        
        JoinIndexPtr JoinIndexes::createJoinIndex(TablePtr pk_tab, std::string pk_tab_col_name, TablePtr fk_tab, std::string fk_tab_col_name){
            TablePtr tab = BaseTable::join(pk_tab, pk_tab_col_name, fk_tab, fk_tab_col_name, HASH_JOIN, LOOKUP, CPU);
//            std::list<std::string> column_names; 
//            column_names.push_back(pk_tab_col_name);
//            column_names.push_back(fk_tab_col_name);
//            tab=BaseTable::sort(tab,column_names, ASCENDING, LOOKUP, CPU);
            
            LookupTablePtr lookup_table = boost::dynamic_pointer_cast<LookupTable>(tab);
            std::vector<LookupColumnPtr> lookup_columns = lookup_table->getLookupColumns();
            assert(lookup_columns.size()==2);
            JoinIndexPtr join_index(new JoinIndex(std::make_pair(lookup_columns[0],lookup_columns[1])));
            
//            
//            std::map<TID,TID> sorted_pairs;
//            PositionListPtr pk_column_tids = join_index->first->getPositionList();
//            PositionListPtr fk_column_tids = join_index->second->getPositionList();
//            for(unsigned int i=0;i<lookup_table->getNumberofRows();++i){
//                sorted_pairs.push_back(TID_Pair(pk_column_tids[i], fk_column_tids[i]));
//            }
            
            
            
            //TODO: FIXME: sort after column 0, then after column 1
            
            typedef std::pair<TID,TID> TID_Pair;
            typedef std::vector<TID_Pair>  TID_Pair_Vector;
            TID_Pair_Vector join_pairs;
            PositionListPtr pk_column_tids = join_index->first->getPositionList();
            PositionListPtr fk_column_tids = join_index->second->getPositionList();
            for(unsigned int i=0;i<lookup_table->getNumberofRows();++i){
                join_pairs.push_back(TID_Pair((*pk_column_tids)[i], (*fk_column_tids)[i]));
            }
            std::sort(join_pairs.begin(), join_pairs.end());
            
            PositionListPtr pk_column_tids_sorted(new PositionList(lookup_table->getNumberofRows()));
            PositionListPtr fk_column_tids_sorted(new PositionList(lookup_table->getNumberofRows()));
            for(unsigned int i=0;i<join_pairs.size();++i){
                (*pk_column_tids_sorted)[i]=join_pairs[i].first;
                (*fk_column_tids_sorted)[i]=join_pairs[i].second;
            }
            
            LookupColumnPtr pk_lc_sorted(new LookupColumn(join_index->first->getTable(),pk_column_tids_sorted)); 
            LookupColumnPtr fk_lc_sorted(new LookupColumn(join_index->second->getTable(),fk_column_tids_sorted)); 
            join_index->first=pk_lc_sorted;
            join_index->second=fk_lc_sorted;
            return join_index;
        }
    
    

    
    //fetch TIDs of FK Column from JoinIndex
    //matching tids comes from a filter operation on a primary key table, and we now seek the corresponding foreign key tids
    //Note: assumes the Join Index is sorted by the first PositionList (the Primary Key Table TID list)
    PositionListPtr fetchMatchingTIDsFromJoinIndex(JoinIndexPtr join_index, PositionListPtr matching_tids){
        assert(join_index!=NULL);
        assert(matching_tids!=NULL);
        
        PositionListPtr fk_column_matching_tids(new PositionList());
        PositionListPtr pk_column_tids = join_index->first->getPositionList();
        PositionListPtr fk_column_tids = join_index->second->getPositionList();
        //intersect(pk_column_matching_tids, matching_tids) -> matching fks
        unsigned int left_pos=0;
        unsigned int right_pos=0;  
        unsigned int number_of_join_pairs=join_index->first->getPositionList()->size();
        //assume that matching tids contains no duplicates and is sorted
        while(left_pos < matching_tids->size() && right_pos < number_of_join_pairs){
            //std::cout << "Left: " << (*matching_tids)[left_pos] << "   Right: " << (*pk_column_tids)[right_pos] << std::endl;
            TID left_value = (*matching_tids)[left_pos];
            TID right_value = (*pk_column_tids)[right_pos];
            if((*matching_tids)[left_pos] == (*pk_column_tids)[right_pos]){
                fk_column_matching_tids->push_back((*fk_column_tids)[right_pos]);
                right_pos++;
            }else if((*matching_tids)[left_pos] < (*pk_column_tids)[right_pos]){
                left_pos++;
            }else if((*matching_tids)[left_pos] > (*pk_column_tids)[right_pos]){
                right_pos++;
            }
        }
        return fk_column_matching_tids;
    }
    
    PositionListPairPtr fetchJoinResultFromJoinIndex(JoinIndexPtr join_index, PositionListPtr matching_tids_pk_column){
        //join result tids
        PositionListPtr pk_column_matching_tids(new PositionList());
        PositionListPtr fk_column_matching_tids(new PositionList());
        PositionListPairPtr join_tids(new PositionListPair(std::make_pair(pk_column_matching_tids,fk_column_matching_tids)));
        
        PositionListPtr pk_column_tids = join_index->first->getPositionList();
        PositionListPtr fk_column_tids = join_index->second->getPositionList();
        //intersect(pk_column_matching_tids, matching_tids) -> matching fks
        unsigned int left_pos=0;
        unsigned int right_pos=0;  
        unsigned int number_of_join_pairs=join_index->first->getPositionList()->size();
        //assume that matching tids contains no duplicates and is sorted
        while(left_pos < matching_tids_pk_column->size() && right_pos < number_of_join_pairs){
            if((*matching_tids_pk_column)[left_pos] == (*pk_column_tids)[right_pos]){
                //found matching join pair
                join_tids->first->push_back((*pk_column_tids)[right_pos]);
                join_tids->second->push_back((*fk_column_tids)[right_pos]);
                right_pos++;
            }else if((*matching_tids_pk_column)[left_pos] < (*pk_column_tids)[right_pos]){
                left_pos++;
            }else if((*matching_tids_pk_column)[left_pos] > (*pk_column_tids)[right_pos]){
                right_pos++;
            }
        }
        return join_tids;
    }
    

    
//    void setBitmapMatchingTIDsFromJoinIndex(JoinIndexPtr join_index, PositionListPtr matching_tids, char* fk_column_matching_rows_bitmap, size_t bitmap_num_bits){
    BitmapPtr createBitmapOfMatchingTIDsFromJoinIndex(JoinIndexPtr join_index, PositionListPtr matching_tids){
        //create bitmap, init every bit with 0
        BitmapPtr result(new Bitmap(join_index->second->getPositionList()->size(), false, true));
        std::cout << "Bitmap Fetch Join: #Rows in Bitmap: " << join_index->second->getPositionList()->size() << std::endl;
        char* fk_column_matching_rows_bitmap = result->data();
        PositionListPtr pk_column_tids = join_index->first->getPositionList();
        //PositionListPtr fk_column_tids = join_index->second->getPositionList();
        //intersect(pk_column_matching_tids, matching_tids) -> matching fks
        unsigned int left_pos=0;
        unsigned int right_pos=0;  
        unsigned int number_of_join_pairs=join_index->first->getPositionList()->size();
        //assume that matching tids contains no duplicates and is sorted
        while(left_pos < matching_tids->size() && right_pos < number_of_join_pairs){
            if((*matching_tids)[left_pos] == (*pk_column_tids)[right_pos]){
                //fk_column_matching_tids->push_back((*fk_column_tids)[right_pos]);
                unsigned int current_bit= right_pos & 7; //i%8;;
                unsigned int bitmask = 1 << current_bit;
                fk_column_matching_rows_bitmap[right_pos/8]|=bitmask;    
                right_pos++;
            }else if((*matching_tids)[left_pos] < (*pk_column_tids)[right_pos]){
                left_pos++;
            }else if((*matching_tids)[left_pos] > (*pk_column_tids)[right_pos]){
                right_pos++;
            }
        }
        return result;
    }
    
    
}; //end namespace CogaDB
