#include <stdio.h>
#include <string>
#include <iostream>
#include <exception>

#include <boost/iostreams/device/file_descriptor.hpp>
#include <boost/iostreams/stream.hpp>

#include "sql/server/sql_parsetree.hpp"
#include "sql/server/sql_driver.hpp"
#include "sql_parser.hpp"

#include <optimizer/optimizer.hpp>

#define NDEBUG

using namespace boost;

namespace CoGaDB {
namespace SQL {

Driver::Driver()
	: buffer(NULL), istream(NULL), parser(*this, scanner)
{
	init_scan();

#ifndef NDEBUG
	parser.set_debug_level(1);
#endif
}

ParseTree::SequencePtr
Driver::parse(std::istream &is)
{
	set_input(is);
	if (parser.parse())
		throw ParseError();

	return result;
}

ParseTree::SequencePtr
Driver::parse(const std::string &src)
{
	set_input(src);
	if (parser.parse())
		throw ParseError();

	return result;
}

Driver::~Driver()
{
	destroy_scan();
}

bool
commandlineExec(const std::string &input)
{
	Timestamp begin = getTimestamp();
	Driver driver;
	ParseTree::SequencePtr seq;
	TablePtr result;

	try {
		seq = driver.parse(input);
	} catch (const std::exception &e) {
		std::cerr << e.what() << std::endl;
		return false;
	}

	result = seq->execute();
	Timestamp end = getTimestamp();
	assert(end >= begin);
	if (result)
		result->print();
        //std::streamsize precision = std::cout.precision();
        //std::cout.precision(3);
        //std::cout << "Execution Time: "; // << double(end-begin)/(1000*1000) << "ms" << std::endl;
        //std::cout.precision(precision);
        printf("Execution Time: %.5lf ms\n", double(end-begin)/(1000*1000));
        //std::cout << std::endl;
	return true;
}

bool
commandlineExplain(const std::string &input)
{
	Driver driver;
	ParseTree::SequencePtr seq;

	try {
		seq = driver.parse(input);
	} catch (const std::exception &e) {
		std::cerr << e.what() << std::endl;
		return false;
	}

	/*
	 * FIXME: this only works on POSIX
	 */
	FILE *pipe = popen("dot -Tgtk", "w");
	if (!pipe) {
		perror("Error opening `dot`");
		return false;
	}
	iostreams::file_descriptor_sink fd(fileno(pipe), iostreams::close_handle);
	iostreams::stream<iostreams::file_descriptor_sink> stream(fd);

	seq->explain(stream);
        
	pclose(pipe);
	return true;
}

bool commandlineExplainStatements(const std::string &input, bool optimize){
 
	Driver driver;
	ParseTree::SequencePtr seq;

	try {
		seq = driver.parse(input);
	} catch (const std::exception &e) {
		std::cerr << e.what() << std::endl;
		return false;
	}
        std::list<CoGaDB::query_processing::LogicalQueryPlanPtr> plans = seq->getLogicalQueryPlans();
        std::list<CoGaDB::query_processing::LogicalQueryPlanPtr>::iterator it;
        unsigned int i=0;
        for(it=plans.begin();it!=plans.end();++it){
            std::cout << "Query Plan for Statement " << i++ <<" : " << std::endl;
            query_processing::LogicalQueryPlanPtr plan=*it;
            if(plan){
                if(optimize){
                    optimizer::Logical_Optimizer::instance().optimize(plan);
                }
                plan->print();
            }    
        }
        return true;
}

bool commandlineExplainStatementsWithoutOptimization(const std::string &input){
    return commandlineExplainStatements(input, false);
}

bool commandlineExplainStatementsWithOptimization(const std::string &input){
    return commandlineExplainStatements(input,true);
}


} /* namespace SQL */
} /* namespace CoGaDB */
