
#include <stdio.h>

#include <iostream>
#include <list>
#include <vector>

#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>

#include <query_processing/query_processor.hpp>
#include <optimizer/optimizer.hpp>
#include <query_processing/query_processor.hpp>

#include "sql/server/sql_parsetree.hpp"

//#define NDEBUG

using namespace CoGaDB;
using namespace query_processing;

namespace CoGaDB {
namespace SQL {
namespace ParseTree {

TablePtr
CreateTable::execute(void)
{
	getGlobalTableList().push_back(table);

	return table;
}

//void
//ScalarExpression::setColumnName(const String &name)
//{
//	std::cerr << "Setting name on column is not supported" << std::endl;
//	/* FIXME */
//        
//}

void
AtomExpression::setColumnName(const String &name)
{
	/* set by AS clause */
	explicit_result_colname = name;
}

String
AtomExpression::toString()
{
	char buf[255];

	/*
	 * Holy god of C++ idiosyncracies, forgive me!
	 */
	switch (type) {
	case INT:
		snprintf(buf, sizeof(buf), "%d",
			 boost::any_cast<int>(atom));
		return String(buf);

	case FLOAT:
		snprintf(buf, sizeof(buf), "%f",
			 boost::any_cast<float>(atom));
		return String(buf);

	case VARCHAR:
		return "'"+boost::any_cast<std::string>(atom)+"'";

	case BOOLEAN:
		return boost::any_cast<bool>(atom)
				? String("TRUE") : String("FALSE");
	}

	/* not reached */
	return String("");
}

String
AtomExpression::getQueryPlan(TypedNodePtr &io_node)
{
	TypedNodePtr result;
	String result_col = explicit_result_colname;

	if (result_col.empty())
		result_col = "("+toString()+")";

	result = boost::make_shared
		 <logical_operator::Logical_AddConstantValueColumn>(result_col, type, atom);
	result->setLeft(io_node);

	io_node = result;
	return result_col;
}

void ColumnExpression::setColumnName(const String &name){
    column=name;
}

String
ColumnExpression::getQueryPlan(TypedNodePtr &io_node)
{
	return column;
}

void
FunctionExpression::setColumnName(const String &name)
{
	return param->setColumnName(name);
}

String
FunctionExpression::getQueryPlan(TypedNodePtr &io_node)
{
	return param->getQueryPlan(io_node);
}

void
AlgebraExpression::setColumnName(const String &name)
{
	/* set by AS clause */
	explicit_result_colname = name;
}

String
AlgebraExpression::getQueryPlan(TypedNodePtr &io_node)
{
	/* FIXME: in C++ this might not be the perfect data structure:
	   what if ColumnAlgebraOperation changes? */
	static const char *op2name[] = {
		/* [ADD] = */ "+",
		/* [SUB] = */ "-",
		/* [MUL] = */ "*",
		/* [DIV] = */ "/"
	};

	String result_col = explicit_result_colname;
	TypedNodePtr result;

	/*
	 * TODO: if both operands are atomic, we do not
	 * need to create two logical operators.
	 * But this optimization has to be performed before
	 * query plan generation (as an algebraic optimization)
	 */
	if (dynamic_cast<AtomExpression *>(lvalue.get()) ||
	    dynamic_cast<AtomExpression *>(rvalue.get())) {
		String col;
		AtomExpressionPtr value;

		if (dynamic_cast<AtomExpression *>(lvalue.get())) {
			value = boost::static_pointer_cast<AtomExpression>(lvalue);
			col = rvalue->getQueryPlan(io_node);
			if (result_col.empty())
				result_col = "("+value->toString()+op2name[op]+col+")";
		} else {
			value = boost::static_pointer_cast<AtomExpression>(rvalue);
			col = lvalue->getQueryPlan(io_node);
			if (result_col.empty())
				result_col = "("+col+op2name[op]+value->toString()+")";
		}
		
		result = boost::make_shared<logical_operator::Logical_ColumnConstantOperator>
								(col, value->atom, result_col, op);
	} else {
		String l_col = lvalue->getQueryPlan(io_node);
		String r_col = rvalue->getQueryPlan(io_node);

		if (result_col.empty())
			result_col = "("+l_col+op2name[op]+r_col+")";

		result = boost::make_shared<logical_operator::Logical_ColumnAlgebraOperator>
								(l_col, r_col, result_col, op);
	}
	result->setLeft(io_node);

	io_node = result;
	return result_col;
}

/*!
 * \brief Convert a parse tree \b AND condition into
 *        Conjunctive Normal Form
 *
 * To convert the \b AND condition of two search conditions,
 * conjunctions of their CNF simply have to be concatenated.
 *
 * \param io_node Reference to node pointer that holds the current
 *                chain of algebraic logical operators that is updated
 *                during CNF construction.
 * \return The CNF of the conjunction of the two search conditions
 */
KNF_Selection_Expression
AndCondition::getCNF(TypedNodePtr &io_node)
{
	KNF_Selection_Expression l_knf = lvalue->getCNF(io_node);
	KNF_Selection_Expression r_knf = rvalue->getCNF(io_node);

	l_knf.disjunctions.insert(l_knf.disjunctions.end(),
				  r_knf.disjunctions.begin(),
				  r_knf.disjunctions.end());

	return l_knf;
}

/*!
 * \brief Convert a parse tree \b OR condition into
 *        Conjunctive Normal Form
 *
 * In an \b OR condition the CNF of lvalue (<tt>lvalue->getCNF()</tt>) and
 * rvalue (<tt>rvalue->getCNF()</tt>) are of the following form: 
 * \remark
 *   L1 \b AND L2 \b AND L3 ... \b AND Ln<br/>
 *   R1 \b AND R2 \b AND R3 ... \b AND Rm
 *
 * The disjunction of these terms is equivalent to:
 * \remark
 *   (L1 \b OR R1) \b AND (L1 \b OR R2) ... \b AND (L1 \b OR Rm) \b AND<br/>
 *   (L2 \b OR R1) \b AND (L2 \b OR R2) ... \b AND (L2 \b OR Rm) \b AND ...<br/>
 *   (Ln \b OR R1) \b AND (Ln \b OR R2) ... \b AND (Ln \b OR Rm)
 *
 * \param io_node Reference to node pointer that holds the current
 *                chain of algebraic logical operators that is updated
 *                during CNF construction.
 * \return The CNF of the disjunction of the two search conditions
 */
KNF_Selection_Expression
OrCondition::getCNF(TypedNodePtr &io_node)
{
	KNF_Selection_Expression l_knf = lvalue->getCNF(io_node);
	KNF_Selection_Expression r_knf = rvalue->getCNF(io_node);
	KNF_Selection_Expression result;

	result.disjunctions.reserve(l_knf.disjunctions.size() *
				    r_knf.disjunctions.size());

	for (int n = 0; n < l_knf.disjunctions.size(); n++) {
		Disjunction &l_disj = l_knf.disjunctions[n];

		for (int m = 0; m < r_knf.disjunctions.size(); m++) {
			Disjunction &r_disj = r_knf.disjunctions[m];
			Disjunction result_disj;

			result_disj.reserve(l_disj.size() + r_disj.size());

			result_disj.insert(result_disj.end(),
					   l_disj.begin(), l_disj.end());
			result_disj.insert(result_disj.end(),
					   r_disj.begin(), r_disj.end());

			result.disjunctions.push_back(result_disj);
		}		
	}

	return result;
}

/*!
 * \brief Construct CoGaDB::Predicate
 *
 * It resolves boost::any to strings for ValueValuePredicates.
 *
 * \param column Name of column to compare
 * \param value Value or name of column to compare with
 * \param pred Predicate type
 * \param comp Comparison operator
 * \return a CoGaDB::Predicate
 */
static CoGaDB::Predicate
make_pred(const std::string &column, const boost::any &value,
	  PredicateType pred, CoGaDB::ValueComparator comp)
{
	if (pred == ValueValuePredicate)
		return CoGaDB::Predicate(column, boost::any_cast<std::string>(value),
					 pred, comp);
	else
		return CoGaDB::Predicate(column, value, pred, comp);
}

/*!
 * \brief Get Conjunctive Normal Form of a search condition predicate
 *
 * Converting a single predicate to CNF is trivial - it is
 * simply wrapped in a disjunction that is wrapped in a conjunction.
 *
 * Since predicates in a CNF structure have their atomic type always
 * on the right side of the operator (as in Column > Atom),
 * the predicate operator might have to be flipped.
 *
 * \see CoGaDB::Predicate
 *
 * \param io_node Reference to node pointer that holds the current
 *                chain of algebraic logical operators that is updated
 *                during CNF construction.
 * \return CNF of parse tree predicate
 */
KNF_Selection_Expression
ComparisonPredicate::getCNF(TypedNodePtr &io_node)
{
	KNF_Selection_Expression knf;
	Disjunction disjunction;

	std::string cogadb_column;
	boost::any cogadb_value; /* may be atom or column name */
	PredicateType cogadb_type;
	ValueComparator normalized_op = op;

	if (!dynamic_cast<AtomExpression *>(lvalue.get()) &&
	    !dynamic_cast<AtomExpression *>(rvalue.get())) {
		cogadb_column = lvalue->getQueryPlan(io_node);
		cogadb_value = boost::any(rvalue->getQueryPlan(io_node));
		cogadb_type = ValueValuePredicate;
	} else if (dynamic_cast<AtomExpression *>(rvalue.get())) {
		AtomExpressionPtr ratom(boost::static_pointer_cast<AtomExpression>(rvalue));

		cogadb_column = lvalue->getQueryPlan(io_node);
		cogadb_value = ratom->atom;
		cogadb_type = ValueConstantPredicate;
	} else if (dynamic_cast<AtomExpression *>(lvalue.get())) {
		AtomExpressionPtr latom(boost::static_pointer_cast<AtomExpression>(lvalue));

		cogadb_column = rvalue->getQueryPlan(io_node);
		cogadb_value = latom->atom;
		cogadb_type = ValueConstantPredicate;

		/* operands are switched, so reverse operator */
		switch (normalized_op) {
		case LESSER:
			normalized_op = GREATER;
			break;
		case LESSER_EQUAL:
			normalized_op = GREATER_EQUAL;
			break;
		case GREATER:
			normalized_op = LESSER;
			break;
		case GREATER_EQUAL:
			normalized_op = LESSER_EQUAL;
			break;
		case EQUAL:
		case UNEQUAL:
			/* nothing to do */
			break;
		}
	}

	/*!
	 * map SQL comparison operators to CoGaDB's limited set of operators.
	 * \todo If we would save the data type in the AtomExpression we
	 *       could optimize expressions like (x >= 3) to (x > 2), saving
	 *       one predicate.
	 */
	switch (normalized_op) {
	case LESSER_EQUAL:
		disjunction.push_back(make_pred(cogadb_column, cogadb_value,
						cogadb_type, CoGaDB::LESSER_EQUAL));
                break;
	case LESSER:
		disjunction.push_back(make_pred(cogadb_column, cogadb_value,
						cogadb_type, CoGaDB::LESSER));
		break;

	case GREATER_EQUAL:
		disjunction.push_back(make_pred(cogadb_column, cogadb_value,
						cogadb_type, CoGaDB::GREATER_EQUAL));
                break;
	case GREATER:
		disjunction.push_back(make_pred(cogadb_column, cogadb_value,
						cogadb_type, CoGaDB::GREATER));
		break;

	case EQUAL:
		disjunction.push_back(make_pred(cogadb_column, cogadb_value,
						cogadb_type, CoGaDB::EQUAL));
		break;

	case UNEQUAL:
		disjunction.push_back(make_pred(cogadb_column, cogadb_value,
						cogadb_type, CoGaDB::LESSER));
		disjunction.push_back(make_pred(cogadb_column, cogadb_value,
						cogadb_type, CoGaDB::GREATER));
		break;
	}

	knf.disjunctions.push_back(disjunction);
	return knf;
}

/*!
 * \brief Construct BETWEEN predicate
 *
 * Currently the BETWEEN predicate is implemented as a conjunction of
 * to predicates using parse tree nodes.
 * With CoGaDB's current set of supported predicate operators, this results in
 * a CNF like "(exp = val1 OR exp > val1) AND (exp = val2 OR exp < val2)".
 *
 * In the future, we might represent BETWEEN predicates as entirely distinct
 * nodes with custom implementations of getCNF().
 * This would allow us to minimize the amount of logical query plan operators
 * for algebraic operations in the predicate.
 *
 * \param _exp The value that should be checked
 * \param _lvalue The lower inclusive value
 * \param _rvalue The upper inclusive value
 * \returns A BETWEEN predicate node
 */
BetweenPredicate::BetweenPredicate(ScalarExpressionPtr _exp,
				   ScalarExpressionPtr _lvalue,
				   ScalarExpressionPtr _rvalue)
{
	ComparisonPredicatePtr lpred = boost::make_shared<ComparisonPredicate>(_exp, GREATER_EQUAL, _lvalue);
	ComparisonPredicatePtr rpred = boost::make_shared<ComparisonPredicate>(_exp, LESSER_EQUAL, _rvalue);
	and_cond = boost::make_shared<AndCondition>(lpred, rpred);
}

/*!
 * \brief Calculate CNF of a BETWEEN predicate
 */
KNF_Selection_Expression
BetweenPredicate::getCNF(TypedNodePtr &io_node)
{
	return and_cond->getCNF(io_node);
}

/*!
 * \brief Construct NOT BETWEEN predicate
 *
 * \see BetweenPredicate
 */
NotBetweenPredicate::NotBetweenPredicate(ScalarExpressionPtr _exp,
					 ScalarExpressionPtr _lvalue,
					 ScalarExpressionPtr _rvalue)
{
	ComparisonPredicatePtr lpred = boost::make_shared<ComparisonPredicate>(_exp, LESSER, _lvalue);
	ComparisonPredicatePtr rpred = boost::make_shared<ComparisonPredicate>(_exp, GREATER, _rvalue);
	or_cond = boost::make_shared<OrCondition>(lpred, rpred);
}

/*!
 * \brief Calculate CNF of a NOT BETWEEN predicate
 */
KNF_Selection_Expression
NotBetweenPredicate::getCNF(TypedNodePtr &io_node)
{
	return or_cond->getCNF(io_node);
}

TypedNodePtr
TableName::getQueryPlan()
{
	return boost::make_shared<logical_operator::Logical_Scan>(table);
}

TypedNodePtr
InnerJoin::getQueryPlan()
{
	TypedNodePtr join;
	std::string join_col1, join_col2;

	ComparisonPredicatePtr predicate(boost::dynamic_pointer_cast<ComparisonPredicate>(condition));

	if (predicate && predicate->op == EQUAL &&
	    dynamic_cast<ColumnExpression *>(predicate->lvalue.get()) &&
	    dynamic_cast<ColumnExpression *>(predicate->rvalue.get())) {
		ColumnExpressionPtr lvalue(boost::static_pointer_cast<ColumnExpression>(predicate->lvalue));
		ColumnExpressionPtr rvalue(boost::static_pointer_cast<ColumnExpression>(predicate->rvalue));

		join_col1 = lvalue->column;
		join_col2 = rvalue->column;
	} else {
		std::cerr << "Unsupported JOIN predicate!" << std::endl;
		/* FIXME */
	}

	join = boost::make_shared<logical_operator::Logical_Join>(join_col1, join_col2);
	join->setLeft(lvalue->getQueryPlan());
	join->setRight(rvalue->getQueryPlan());

	return join;
}

TypedNodePtr
CrossJoin::getQueryPlan()
{
	TypedNodePtr join = boost::make_shared<logical_operator::Logical_CrossJoin>();
	join->setLeft(lvalue->getQueryPlan());
	join->setRight(rvalue->getQueryPlan());

	return join;
}

TablePtr
InsertInto::execute(void)
{
	TablePtr table_p = getTablebyName(table);

	table_p->insert(row);

	return table_p;
}

TableExpression::TableExpression(TableReferenceList &from, SearchConditionPtr where,
				 ColumnListPtr _group_by, OrderByPtr _order_by)
				: group_by(_group_by), order_by(_order_by)
{
	/*
	 * Compile list of table references into chained logical
	 * cross joins
	 */
	for (TableReferenceList::iterator it = from.begin();
	     it != from.end(); it++) {
		TypedNodePtr plan = (*it)->getQueryPlan();

		if (!table.get()) {
			/* first table reference */
			table = plan;
		} else {
			TypedNodePtr last_table = table;

			table = boost::make_shared<logical_operator::Logical_CrossJoin>();
			table->setLeft(last_table);
			table->setRight(plan);
		}
	}

	/*
	 * FIXME: resolve selections and cross joins to
	 * inner joins.
	 * This can only be done generically at execution time
	 * to decide which column referenced in the predicates
	 * belongs to which table.
	 */
	if (where.get()) {
		/*
		 * FIXME: nodes added to io_node by getCNF() and in turn
		 * ScalarExpression::getQueryPlan() must be projected
		 * out again.
		 * E.g.: SELECT * FROM table WHERE x*2 > 2;
		 * This requires table introspection and keeping a list
		 * of columns added by column operations to io_node.
		 */
		TypedNodePtr io_node = table; /* joins */
		KNF_Selection_Expression knf = where->getCNF(io_node);

		table = boost::make_shared<logical_operator::Logical_ComplexSelection>(knf,LOOKUP,CoGaDB::RuntimeConfiguration::instance().getGlobalDeviceConstraint());
		table->setLeft(io_node);
	}
}

SelectFrom::SelectFrom(ScalarExpressionListPtr columns, TableExpression &table_exp)
{
	ColumnList proj_list;
	std::list<ColumnAggregation> aggr_list;

	node = table_exp.table;

	if (columns) {
		/*
		 * add nodes to io_node required for calculations (like algebraic
		 * expressions) in the column list and aggregate the column names
		 * necessary for projection (since they might be calculated as well).
		 * ColumnAggregation tuples are calculated from function names
		 * and the column names of their arguments.
		 */
		for (ScalarExpressionList::iterator it = columns->begin();
		     it != columns->end(); it++) {
			ScalarExpressionPtr cur = *it;
			String cur_colname = cur->getQueryPlan(node);

			proj_list.push_back(cur_colname);

			if (dynamic_cast<FunctionExpression *>(cur.get())) {
				FunctionExpressionPtr fnc(boost::static_pointer_cast<FunctionExpression>(cur));
				aggr_list.push_back(ColumnAggregation(cur_colname, fnc->function));
			}
		}
	}

	if (table_exp.group_by) {
		TypedNodePtr sort;

		if (aggr_list.empty()) {
			std::cerr << "Invalid GROUP BY!" << std::endl;
			/* FIXME */
		}

		/*
		 * presort table to guarantee that grouping works
		 */
		sort = boost::make_shared<logical_operator::Logical_Sort>
						(*table_exp.group_by, ASCENDING);
		sort->setLeft(node);

		node = boost::make_shared<logical_operator::Logical_Groupby>
						(*table_exp.group_by, aggr_list);
		node->setLeft(sort);
		/*
		 * no additional projection necessary:
		 * GroupBy already discared all non-grouping and non-aggregated columns
		 */
	} else if (!aggr_list.empty()) {
		TypedNodePtr grouping_column;
		std::list<std::string> group(1, "_GROUPING_COLUMN");

		grouping_column = boost::make_shared<logical_operator::Logical_AddConstantValueColumn>
									(group.front(), INT, boost::any(23));
		grouping_column->setLeft(node);

		node = boost::make_shared<logical_operator::Logical_Groupby>(group, aggr_list);
		node->setLeft(grouping_column);

		/* grouping column is later projected out */
	}

	/* try to avoid Sort operation if presorting for Groupy is sufficient */
	if (table_exp.order_by &&
	    !(table_exp.group_by && *table_exp.group_by == table_exp.order_by->list &&
	      table_exp.order_by->order == ASCENDING)) {
		TypedNodePtr sort;

		/*
		 * FIXME: sort order may be different for different columns
		 */
		sort = boost::make_shared<logical_operator::Logical_Sort>
						(table_exp.order_by->list, table_exp.order_by->order);
		sort->setLeft(node);
		node = sort;
	}

	if (!proj_list.empty() && !table_exp.group_by) {
		TypedNodePtr source = node;

		node = boost::make_shared<logical_operator::Logical_Projection>(proj_list);
		node->setLeft(source);
	}
}

TablePtr
SelectFrom::execute(void)
{
	LogicalQueryPlanPtr log_plan(boost::make_shared<LogicalQueryPlan>(node));
        
	optimizer::Logical_Optimizer::instance().optimize(log_plan);
        TablePtr result = CoGaDB::query_processing::optimize_and_execute("",*log_plan);;
        if(result){
            //is result, so set empty string as table name
            result->setName("");
        }
  
        return result;
        
//	PhysicalQueryPlanPtr physical_plan(log_plan->convertToPhysicalQueryPlan());


        
//#ifndef NDEBUG
//	log_plan->print();
//	physical_plan->print();
//#endif
//        
//	physical_plan->run();

	//return physical_plan->getResult();
}

void
SelectFrom::explain(std::ostream &os)
{
	LogicalQueryPlanPtr log_plan(boost::make_shared<LogicalQueryPlan>(node));

	optimizer::Logical_Optimizer::instance().optimize(log_plan);

	log_plan->print_graph(os);
}

CoGaDB::query_processing::LogicalQueryPlanPtr SelectFrom::getLogicalQueryPlan(){
    return CoGaDB::query_processing::LogicalQueryPlanPtr(boost::make_shared<LogicalQueryPlan>(node));
}

TablePtr
Sequence::execute(void)
{
	TablePtr result;

	for (std::list<StatementPtr>::iterator it = statements.begin();
	     it != statements.end();
	     it++) {
		StatementPtr statement = *it;

		result = statement->execute();
	}

	return result;
}

void
Sequence::explain(std::ostream &os)
{
	for (std::list<StatementPtr>::iterator it = statements.begin();
	     it != statements.end();
	     it++) {
		StatementPtr statement = *it;

		statement->explain(os);
	}
}

std::list<CoGaDB::query_processing::LogicalQueryPlanPtr> Sequence::getLogicalQueryPlans()
{
	std::list<CoGaDB::query_processing::LogicalQueryPlanPtr> result_plans;
	for (std::list<StatementPtr>::iterator it = statements.begin();
	     it != statements.end();
	     it++) {
		StatementPtr statement = *it;

		//statement->explain(os);
                result_plans.push_back(statement->getLogicalQueryPlan());
	}
        return result_plans;
}

} /* namespace ParseTree */
} /* namespace SQL */
} /* namespace CoGaDB */
